from typing import Annotated

from pydantic import BaseModel, Field, model_validator
from pydantic_extra_types.country import CountryAlpha2
from pydantic_extra_types.currency_code import Currency

from pyrevolut.api.common import EnumProfileType
from pyrevolut.api.counterparties.resources import ResourceCounterparty


class CreateCounterparty:
    """
    Create a new counterparty to transact with.

    In the Sandbox environment, you cannot add real people and businesses as Revolut counterparties. To help you simulate Create a counterparty requests for counterparties of profile type personal, we created some test accounts. Look inside for test Revtags.

    To add a counterparty via Revtag, use one of these pairs for the name and revtag fields respectively:

        Test User 1 & john1pvki
        Test User 2 & john2pvki
        ...
        Test User 9 & john9pvki
    """

    ROUTE = "/1.0/counterparty"

    class Body(BaseModel):
        """Request body for the endpoint."""

        class ModelIndividualName(BaseModel):
            """
            The name of the individual counterparty.
            Use when company_name isn't specified.
            """

            first_name: Annotated[
                str | None,
                Field(description="The first name of the individual counterparty."),
            ] = None
            last_name: Annotated[
                str | None,
                Field(description="The last name of the individual counterparty."),
            ] = None

        class ModelAddress(BaseModel):
            """
            The address of the counterparty.
            """

            street_line1: Annotated[
                str | None,
                Field(description="Street line 1 information."),
            ] = None
            street_line2: Annotated[
                str | None,
                Field(description="Street line 2 information."),
            ] = None
            region: Annotated[
                str | None,
                Field(description="The name of the region."),
            ] = None
            city: Annotated[
                str | None,
                Field(description="The name of the city."),
            ] = None
            country: Annotated[
                CountryAlpha2,
                Field(
                    description="The country of the counterparty as the 2-letter ISO 3166 code."
                ),
            ]
            postcode: Annotated[
                str,
                Field(description="The postcode of the counterparty address."),
            ]

        company_name: Annotated[
            str | None,
            Field(
                description="""
                The name of the company counterparty. 
                Use when individual_name or name isn't specified and profile_type is business.
                """
            ),
        ] = None
        profile_type: Annotated[
            EnumProfileType | None,
            Field(
                description="""
                The type of the Revolut profile. Used when adding an existing Revolut user via Revtag.
                """
            ),
        ] = None
        name: Annotated[
            str | None,
            Field(
                description="""
                The name of the counterparty that you create for an existing Revolut user via Revtag. 
                Provide the value only when you specify personal for profile_type.                
                """
            ),
        ] = None
        individual_name: Annotated[
            ModelIndividualName | None,
            Field(
                description="""
                The name of the individual counterparty. 
                Use when company_name isn't specified.
                """
            ),
        ] = None
        bank_country: Annotated[
            CountryAlpha2 | None,
            Field(description="The country of the bank as the 2-letter ISO 3166 code."),
        ] = None
        currency: Annotated[
            Currency | None,
            Field(description="ISO 4217 currency code in upper case."),
        ] = None
        revtag: Annotated[
            str | None,
            Field(
                description="""
                The Revtag of the counterparty to add.                
                """
            ),
        ] = None
        account_no: Annotated[
            str | None,
            Field(
                description="""
                The bank account number of the counterparty.               
                """
            ),
        ] = None
        iban: Annotated[
            str | None,
            Field(
                description="""
                The IBAN number of the counterparty's account. This field is displayed for IBAN countries.               
                """
            ),
        ] = None
        sort_code: Annotated[
            str | None,
            Field(
                description="""
                The sort code of the counterparty's account. This field is required for GBP accounts.           
                """
            ),
        ] = None
        routing_number: Annotated[
            str | None,
            Field(
                description="""
                The routing number of the counterparty's account. This field is required for USD accounts.               
                """
            ),
        ] = None
        bic: Annotated[
            str | None,
            Field(
                description="""
                The BIC number of the counterparty's account. This field is required for non-SEPA IBAN/SWIFT.             
                """
            ),
        ] = None
        clabe: Annotated[
            str | None,
            Field(
                description="""
                The CLABE number of the counterparty's account. This field is required for SWIFT MX.               
                """
            ),
        ] = None
        isfc: Annotated[
            str | None,
            Field(
                description="""
                The ISFC number of the counterparty's account. This field is required for INR accounts.               
                """
            ),
        ] = None
        bsb_code: Annotated[
            str | None,
            Field(
                description="""
                The BSB code of the counterparty's account. This field is required for AUD accounts.               
                """
            ),
        ] = None
        address: Annotated[
            ModelAddress | None,
            Field(description="The address of the counterparty."),
        ] = None

        @model_validator(mode="after")
        def check_inputs(self) -> "CreateCounterparty.Body":
            """Validate the input data."""

            # Company name check
            if self.profile_type == EnumProfileType.BUSINESS and (
                self.individual_name is None or self.name is None
            ):
                assert (
                    self.company_name is not None
                ), "company_name is required for business profile type."

            # Name check
            if self.profile_type == EnumProfileType.PERSONAL:
                assert (
                    self.name is not None
                ), "name is required for personal profile type."

            # Profile type check
            if self.profile_type is not None:
                assert (
                    self.revtag is not None
                ), "revtag is required when profile_type is specified."
            else:
                assert (
                    self.revtag is None
                ), "revtag is not required when profile_type is not specified."

            # Individual name check
            # if self.company_name is None:
            #     assert (
            #         self.individual_name is not None
            #     ), "individual_name is required when company_name is not specified."

            # Sort code check
            if self.currency == "GBP":
                assert (
                    self.sort_code is not None
                ), "sort_code is required for GBP accounts."

            # Routing number check
            if self.currency == "USD":
                assert (
                    self.routing_number is not None
                ), "routing_number is required for USD accounts."

            # IFSC check
            if self.currency == "INR":
                assert self.isfc is not None, "isfc is required for INR accounts."

            # BSB code check
            if self.currency == "AUD":
                assert (
                    self.bsb_code is not None
                ), "bsb_code is required for AUD accounts."

            return self

    class Response(ResourceCounterparty):
        """Response model for the endpoint."""

        pass



from uuid import UUID
from typing import Literal

from pyrevolut.exceptions import PyRevolutInvalidEnvironment
from pyrevolut.api.common import (
    BaseEndpointSync,
    EnumTransactionState,
)
from pyrevolut.api.simulations.post import (
    SimulateAccountTopup,
    SimulateTransferStateUpdate,
)


class EndpointSimulationsSync(BaseEndpointSync):
    """The Simulations API

    The Simulations API is only available in the Sandbox environment.
    It lets you simulate certain events that are otherwise only possible in the production environment,
    such as your account's top-up and transfer state changes.
    """

    def simulate_account_topup(
        self,
        account_id: UUID,
        amount: float,
        currency: str,
        reference: str | None = None,
        state: EnumTransactionState | None = None,
        **kwargs,
    ) -> dict | SimulateAccountTopup.Response:
        """
        Simulate a top-up of your account in the Sandbox environment.

        This is useful during testing, when you run out of money in your test account
        and need to add more.

        Parameters
        ----------
        account_id : UUID
            The ID of the account that you want to top up.
        amount : float
            The amount with which you want to top up the account. Must be <= 10000
        currency : str
            The currency of the top-up amount. Must be a valid ISO 4217 currency code.
        reference : str, optional
            A short description for your top up.
            Default value: 'Test Top-up' if not provided.
        state : EnumTransactionState, optional
            The state to which you want to set the top-up transaction.

            If not provided, the default value is 'completed'.

            Possible values:

                pending:
                    The transaction is pending until it's being processed.
                    If the transfer is made between Revolut accounts,
                    this state is skipped and the transaction is executed instantly.
                completed:
                    The transaction was successful.
                failed:
                    The transaction was unsuccessful. This can happen for a variety of reasons,
                    for example, invalid API calls, blocked payments, etc.
                reverted:
                    The transaction was reverted. This can happen for a variety of reasons,
                    for example, the receiver being inaccessible.

        Returns
        -------
        dict | SimulateAccountTopup.Response
            The top-up transaction information.
        """
        self.__check_sandbox()
        endpoint = SimulateAccountTopup
        path = endpoint.ROUTE
        body = endpoint.Body(
            account_id=account_id,
            amount=amount,
            currency=currency,
            reference=reference,
            state=state,
        )

        return self.client.post(
            path=path,
            response_model=endpoint.Response,
            body=body,
            **kwargs,
        )

    def simulate_transfer_state_update(
        self,
        transfer_id: UUID,
        action: Literal["complete", "revert", "decline", "fail"],
        **kwargs,
    ) -> dict | SimulateTransferStateUpdate.Response:
        """
        Simulate a transfer state change in the Sandbox environment.

        For example, after you make a transfer in Sandbox, you can change its
        state to completed.

        The resulting state is final and cannot be changed.

        Parameters
        ----------
        transfer_id : UUID
            The ID of the transfer whose state you want to update.
        action : Literal["complete", "revert", "decline", "fail"]
            The action you want to perform on the transfer. Possible values:

                complete:
                    Simulate a completed transfer.
                revert:
                    Simulate a reverted transfer.
                decline:
                    Simulate a declined transfer.
                fail:
                    Simulate a failed transfer.

        Returns
        -------
        dict | SimulateTransferStateUpdate.Response
            The updated transfer information.
        """
        self.__check_sandbox()
        assert action in ["complete", "revert", "decline", "fail"], "Invalid action"
        endpoint = SimulateTransferStateUpdate
        path = endpoint.ROUTE.format(transfer_id=transfer_id, action=action)
        body = endpoint.Body()

        return self.client.post(
            path=path,
            response_model=endpoint.Response,
            body=body,
            **kwargs,
        )

    def __check_sandbox(self):
        """
        Check if the sandbox is enabled.

        Raises
        ------
        PyRevolutInvalidEnvironment
            If the sandbox is enabled.
        """
        if not self.client.sandbox:
            raise PyRevolutInvalidEnvironment(
                "This feature is only available in the Sandbox."
            )



from typing import Annotated
from uuid import UUID

from pydantic import BaseModel, Field, HttpUrl
from pydantic_extra_types.currency_code import Currency

from pyrevolut.utils import DateTime
from pyrevolut.api.common import (
    EnumPayoutLinkState,
    EnumPayoutLinkPaymentMethod,
    EnumTransferReasonCode,
    EnumPayoutLinkCancellationReason,
)


class ResourcePayoutLink(BaseModel):
    """
    Payout Link resource model.
    """

    id: Annotated[UUID, Field(description="The ID of the payout link.")]
    state: Annotated[
        EnumPayoutLinkState,
        Field(
            description="""
            The state that the payout link is in. Possible states are:

                created:
                    The payout link has been created, but the amount has not yet been blocked.
                failed:
                    The payout link couldn't be generated due to a failure during transaction booking.
                awaiting:
                    The payout link is awaiting approval.
                active:
                    The payout link can be redeemed.
                expired:
                    The payout link cannot be redeemed because it wasn't claimed before its expiry date.
                cancelled:
                    The payout link cannot be redeemed because it was cancelled.
                processing:
                    The payout link has been redeemed and is being processed.
                processed:
                    The payout link has been redeemed and the money has been transferred to the recipient.
            """
        ),
    ]
    created_at: Annotated[
        DateTime,
        Field(
            description="The date and time the payout link was created in ISO 8601 format."
        ),
    ]
    updated_at: Annotated[
        DateTime,
        Field(
            description="The date and time the payout link was last updated in ISO 8601 format."
        ),
    ]
    counterparty_name: Annotated[
        str, Field(description="The name of the counterparty provided by the sender.")
    ]
    counterparty_id: Annotated[
        UUID | None,
        Field(
            description="""
            The ID of the counterparty created based on the recipient's details.
            
            By default, the newly created counterparty is hidden from your counterparties list.
            To automatically save it when the link is claimed, pass the save_counterparty parameter set to true.
            Alternatively, you can add the recipient to your counterparties later from the list of transactions 
            in the Business app.
            """
        ),
    ] = None
    save_counterparty: Annotated[
        bool,
        Field(
            description="""
            Indicates whether you chose to save the recipient as your counterparty upon link claim. 
            If false then the counterparty will not show up on your counterparties list, 
            for example, when you retrieve your counterparties. 
            However, you can still retrieve this counterparty by its ID.

            If you didn't choose to save the counterparty on link creation, you can still do it 
            from your transactions list in the Business app.
            """
        ),
    ]
    request_id: Annotated[
        str,
        Field(
            description="The ID of the request, provided by the sender.",
        ),
    ]
    expiry_date: Annotated[
        DateTime,
        Field(description="The datetime the payout link expires in ISO 8601 format."),
    ]
    payout_methods: Annotated[
        list[EnumPayoutLinkPaymentMethod],
        Field(
            description="""
            The list of payout methods that the recipient can use to claim the payout, where:

                revolut:
                    Revolut peer-to-peer (P2P) transfer

                bank_account:
                    External bank transfer

                card:
                    Card transfer
            """
        ),
    ]
    account_id: Annotated[UUID, Field(description="The ID of the sender's account.")]
    amount: Annotated[
        float,
        Field(
            description="""
            The amount of money to be transferred.
            The amount must be between £1 and £2,500, or equivalent in the selected currency.
            """
        ),
    ]
    currency: Annotated[
        Currency,
        Field(description="ISO 4217 currency code in upper case."),
    ]
    transaction_id: Annotated[
        UUID | None,
        Field(
            description="""
            The ID of the created transaction. Returned only if the payout has been claimed.
            """
        ),
    ] = None
    url: Annotated[
        HttpUrl | None,
        Field(
            description="The URL of the payout link. Returned only for active payout links."
        ),
    ] = None
    reference: Annotated[
        str,
        Field(
            description="The reference for the payout transaction, provided by the sender."
        ),
    ]
    transfer_reason_code: Annotated[
        EnumTransferReasonCode | None,
        Field(
            description="""
            The reason code for the transaction. 
            Transactions to certain countries and currencies might require you to 
            provide a transfer reason. 
            You can check available reason codes with the getTransferReasons operation.

            If a transfer reason is not required for the given currency and country, 
            this field is ignored.
            """
        ),
    ] = None
    cancellation_reason: Annotated[
        EnumPayoutLinkCancellationReason | None,
        Field(
            description="""
            The reason why the payout link was cancelled. Possible reasons are:

                too_many_name_check_attempts:
                    The name check failed too many times.
            """
        ),
    ] = None



"""This module holds all POST transfers endpoints."""

# flake8: noqa: F401
from .create_transfer_to_another_account import CreateTransferToAnotherAccount
from .move_money_between_accounts import MoveMoneyBetweenAccounts



from enum import StrEnum


class EnumTeamMemberState(StrEnum):
    """Enum for team member state"""

    CREATED = "created"
    CONFIRMED = "confirmed"
    WAITING = "waiting"
    ACTIVE = "active"
    LOCKED = "locked"
    DISABLED = "disabled"



from pydantic import BaseModel


class DeleteCounterparty:
    """Delete a counterparty with the given ID.
    When a counterparty is deleted, you cannot make any payments to the counterparty.
    """

    ROUTE = "/1.0/counterparty/{counterparty_id}"

    class Params(BaseModel):
        """
        Query parameters for the endpoint.
        """

        pass

    class Response(BaseModel):
        """
        Response model for the endpoint.
        """

        pass



from enum import StrEnum


class EnumTransactionType(StrEnum):
    """Transaction type enum."""

    ATM = "atm"
    CARD_PAYMENT = "card_payment"
    CARD_REFUND = "card_refund"
    CARD_CHARGEBACK = "card_chargeback"
    CARD_CREDIT = "card_credit"
    EXCHANGE = "exchange"
    TRANSFER = "transfer"
    LOAN = "loan"
    FEE = "fee"
    REFUND = "refund"
    TOPUP = "topup"
    TOPUP_RETURN = "topup_return"
    TAX = "tax"
    TAX_REFUND = "tax_refund"
    TEMP_BLOCK = "temp_block"



"""This module holds all the enums used in the API."""

# flake8: noqa: F401
from .account_name_match_code import EnumAccountNameMatchCode
from .account_name_match_reason_code import EnumAccountNameMatchReasonCode
from .account_name_match_reason_type import EnumAccountNameMatchReasonType
from .account_state import EnumAccountState
from .account_type import EnumAccountType
from .card_scheme import EnumCardScheme
from .card_state import EnumCardState
from .charge_bearer import EnumChargeBearer
from .merchant_category import EnumMerchantCategory
from .payment_draft_state import EnumPaymentDraftState
from .payment_scheme import EnumPaymentScheme
from .payout_link_cancellation_reason import EnumPayoutLinkCancellationReason
from .payout_link_payment_method import EnumPayoutLinkPaymentMethod
from .payout_link_state import EnumPayoutLinkState
from .profile_state import EnumProfileState
from .profile_type import EnumProfileType
from .recipient_charges import EnumRecipientCharges
from .team_member_state import EnumTeamMemberState
from .time_unit import EnumTimeUnit
from .transaction_type import EnumTransactionType
from .transaction_state import EnumTransactionState
from .transfer_reason_code import EnumTransferReasonCode
from .webhook_event import EnumWebhookEvent



class PyRevolutBaseException(Exception):
    """Base exception for all pyrevolut exceptions."""

    def __init__(
        self,
        msg="An error occurred while processing your request to Revolut.",
        *args,
        **kwargs,
    ):
        super().__init__(msg, *args, **kwargs)



"""Manage counterparties that you intend to transact with.

Request and response examples can vary based on the account provider's
location and type of the counterparty.

In the Sandbox environment, you cannot add real people and businesses as Revolut counterparties.
Therefore, to help you simulate Create a counterparty requests, we have created some
test accounts for counterparties of profile type personal.

To add a counterparty via Revtag, use one of these pairs for the name and revtag fields respectively:

Test User 1 & john1pvki
Test User 2 & john2pvki
...
Test User 9 & john9pvki

Reference link: https://developer.revolut.com/docs/business/counterparties
"""

# flake8: noqa: F401
from .endpoint import EndpointCounterpartiesSync, EndpointCounterpartiesAsync



from datetime import datetime

import jwt

from pyrevolut.utils.datetime import DateTime, to_datetime


def create_client_assert_jwt(
    client_id: str,
    expiration_dt: datetime | DateTime | str | int | float,
    private_credentials_key: bytes,
    issuer_url: str = "https://example.com",
    save_location: str | None = "client_assertion.jwt",
):
    """
    Method to create a client-assertion JWT for Revolut.
    The JWT is used to authenticate the client to the Revolut API.

    Parameters
    ----------
    client_id : str
        The client ID.
        This is the client ID provided by Revolut when you upload your public key.
    expiration_dt : datetime | DateTime | str | int | float
        The expiration datetime (UTC) of the JWT.
        This can be a datetime object, a DateTime object, a string, an integer, or a float.
    private_credentials_key : bytes
        The private key used to sign the JWT.
    issuer_url : str, optional
        The issuer URL.
        This is the URL of the client.
        When you finalize the authentication using the JWT, Revolut will redirect you to this URL
        with an attached authorization code in the query string.

        For example if the URL is "https://example.com",
        the redirect URL will be https://example.com?code=oa_prod_vYo3mAI9TmJuo2_ukYlHVZMh3OiszmfQdgVqk_gLSkU

        Default is "https://example.com".
    save_location : str | None, optional
        The location to save the JWT assertion.
        If None, the assertion will not be saved.
        Default is "client_assertion.jwt".

    Returns
    -------
    str
        The JWT assertion string.
    """

    # Remove the https:// from the issuer_url if it exists
    issuer_url = issuer_url.replace("https://", "")

    # Convert the expiration_dt to a UNIX timestamp if it is not already
    expiration_dt = to_datetime(dt=expiration_dt)
    expiration_ts = expiration_dt.int_timestamp

    # Create the JWT payload
    payload = {
        "iss": issuer_url,  # The URL of the client
        "sub": client_id,  # The client ID
        "aud": "https://revolut.com",  # The URL of the authorization server
        "exp": expiration_ts,  # UNIX timestamp (integer)
    }

    # Create the JWT assertion using the private key
    assertion_string = jwt.encode(
        payload=payload, key=private_credentials_key, algorithm="RS256"
    )

    # Save the JWT assertion to a file if save_location is provided
    if save_location is not None:
        with open(save_location, "w") as f:
            f.write(assertion_string)

    # Return the JWT assertion
    return assertion_string



"""Get the details of your transactions.

Note
----
An incoming or outgoing payment is represented as a transaction.

Reference link: https://developer.revolut.com/docs/business/team-members
"""

# flake8: noqa: F401
from .endpoint import EndpointTransactionsSync, EndpointTransactionsAsync



from typing import Annotated
from uuid import UUID

from pydantic import BaseModel, Field
from pydantic_extra_types.currency_code import Currency
from pydantic_extra_types.country import CountryAlpha2

from pyrevolut.utils import DateTime
from pyrevolut.api.common import (
    EnumProfileType,
    EnumProfileState,
    EnumAccountType,
    EnumRecipientCharges,
    EnumCardScheme,
)


class ResourceCounterparty(BaseModel):
    """
    Counterparty resource model.
    """

    class ModelAccount(BaseModel):
        """A public account associated with this counterparty."""

        id: Annotated[
            UUID,
            Field(description="The ID of the counterparty's account."),
        ]
        name: Annotated[
            str | None,
            Field(description="The name of the counterparty."),
        ] = None
        bank_country: Annotated[
            CountryAlpha2 | None,
            Field(description="The country of the bank as the 2-letter ISO 3166 code."),
        ]
        currency: Annotated[
            Currency,
            Field(description="ISO 4217 currency code in upper case."),
        ]
        type: Annotated[
            EnumAccountType,
            Field(description="Indicates the type of account."),
        ]
        account_no: Annotated[
            str | None,
            Field(description="The bank account number of the counterparty."),
        ] = None
        iban: Annotated[
            str | None,
            Field(
                description="The IBAN number of the counterparty's account if applicable."
            ),
        ] = None
        sort_code: Annotated[
            str | None,
            Field(
                description="The sort code of the counterparty's account if applicable."
            ),
        ] = None
        routing_number: Annotated[
            str | None,
            Field(
                description="The routing number of the counterparty's account if applicable."
            ),
        ] = None
        bic: Annotated[
            str | None,
            Field(
                description="The BIC number of the counterparty's account if applicable."
            ),
        ] = None
        clabe: Annotated[
            str | None,
            Field(
                description="The CLABE number of the counterparty's account if applicable."
            ),
        ] = None
        ifsc: Annotated[
            str | None,
            Field(
                description="The IFSC number of the counterparty's account if applicable."
            ),
        ] = None
        bsb_code: Annotated[
            str | None,
            Field(
                description="The BSB code of the counterparty's account if applicable."
            ),
        ] = None
        recipient_charges: Annotated[
            EnumRecipientCharges | None,
            Field(description="Indicates the possibility of the recipient charges."),
        ] = None

    class ModelCard(BaseModel):
        """The list of cards associated with this counterparty."""

        id: Annotated[
            UUID,
            Field(description="The ID of the counterparty's card."),
        ]
        name: Annotated[
            str,
            Field(description="The name of the counterparty."),
        ]
        last_digits: Annotated[
            str,
            Field(description="The last four digits of the card number."),
        ]
        scheme: Annotated[
            EnumCardScheme,
            Field(description="The card brand."),
        ]
        country: Annotated[
            CountryAlpha2,
            Field(
                description="The country of the card issuer as the 2-letter ISO 3166 code."
            ),
        ]
        currency: Annotated[
            Currency,
            Field(description="ISO 4217 currency code in upper case."),
        ]

    id: Annotated[
        UUID,
        Field(description="The ID of the counterparty."),
    ]
    name: Annotated[
        str,
        Field(description="The name of the counterparty."),
    ]
    revtag: Annotated[
        str | None,
        Field(description="The Revtag of the counterparty."),
    ] = None
    profile_type: Annotated[
        EnumProfileType | None,
        Field(
            description="The type of the Revolut profile. Used when adding an existing Revolut user via Revtag."
        ),
    ] = None
    country: Annotated[
        CountryAlpha2 | None,
        Field(
            description="The bank country of the counterparty as the 2-letter ISO 3166 code."
        ),
    ] = None
    state: Annotated[
        EnumProfileState,
        Field(description="Indicates the state of the counterparty."),
    ]
    created_at: Annotated[
        DateTime,
        Field(
            description="The date and time the counterparty was created in ISO 8601 format."
        ),
    ]
    updated_at: Annotated[
        DateTime,
        Field(
            description="The date and time the counterparty was last updated in ISO 8601 format."
        ),
    ]
    accounts: Annotated[
        list[ModelAccount] | None,
        Field(
            description="The list of public accounts associated with this counterparty."
        ),
    ] = None
    cards: Annotated[
        list[ModelCard] | None,
        Field(description="The list of cards associated with this counterparty."),
    ] = None



from typing import Annotated
from uuid import UUID

from pydantic import BaseModel, Field, HttpUrl

from pyrevolut.api.common import EnumWebhookEvent


class ResourceWebhook(BaseModel):
    """
    Webhook resource model.
    """

    id: Annotated[UUID, Field(description="The ID of the webhook.")]
    url: Annotated[
        HttpUrl,
        Field(
            description="The valid webhook URL that event notifications are sent to. The supported protocol is https."
        ),
    ]
    events: Annotated[
        list[EnumWebhookEvent],
        Field(description="The list of events that the webhook is subscribed to."),
    ]



from typing import Annotated

from pydantic import BaseModel, Field

from pyrevolut.utils import DateTime
from pyrevolut.api.common import EnumTransactionState


class ResourceTransfer(BaseModel):
    """
    Transfer resource model.
    """

    id: Annotated[str, Field(description="The ID of the transaction created.")]
    state: Annotated[
        EnumTransactionState,
        Field(
            description="""
            Indicates the transaction state. Possible values:

            created:
                The transaction has been created and is either processed asynchronously
                or scheduled for a later time.
            pending:
                The transaction is pending until it's being processed.
                If the transfer is made between Revolut accounts,
                this state is skipped and the transaction is executed instantly.
            completed:
                The transaction was successful.
            declined:
                The transaction was unsuccessful. This can happen for a variety of reasons,
                for example, insufficient account balance, wrong receiver information, etc.
            failed:
                The transaction was unsuccessful. This can happen for a variety of reasons,
                for example, invalid API calls, blocked payments, etc.
            reverted:
                The transaction was reverted. This can happen for a variety of reasons,
                for example, the receiver being inaccessible.
            """
        ),
    ]
    created_at: Annotated[
        DateTime,
        Field(
            description="The date and time the transaction was created in ISO 8601 format."
        ),
    ]
    completed_at: Annotated[
        DateTime | None,
        Field(
            description="The date and time the transaction was completed in ISO 8601 format."
        ),
    ] = None



from .common import PyRevolutBaseException


class PyRevolutNotFound(PyRevolutBaseException):
    """Not Found -- The requested resource could not be found."""

    def __init__(
        self,
        msg="The requested resource could not be found on the Revolut server.",
        *args,
        **kwargs,
    ):
        super().__init__(msg, *args, **kwargs)



from typing import Annotated
from uuid import UUID

from pydantic import BaseModel, Field, EmailStr

from pyrevolut.utils import DateTime


class InviteTeamMember:
    """
    Invite a new member to your business account.

    When you invite a new team member to your business account,
    an invitation is sent to their email address that you provided in this request.
    To join your business account, the new team member has to accept this invitation.

    Note
    ----
    This feature is available in the UK, US and the EEA.

    This feature is not available in Sandbox.
    """

    ROUTE = "/1.0/team-members"

    class Body(BaseModel):
        """
        The body of the request.
        """

        email: Annotated[
            EmailStr, Field(description="The email address of the invited member.")
        ]
        role_id: Annotated[
            UUID | str,
            Field(description="The ID of the role to assign to the new member."),
        ]

    class Response(BaseModel):
        """The response model."""

        email: Annotated[
            EmailStr, Field(description="The email address of the invited member.")
        ]
        id: Annotated[UUID, Field(description="The ID of the invited member.")]
        role_id: Annotated[
            UUID | str,
            Field(description="The ID of the role assigned to the member."),
        ]
        created_at: Annotated[
            DateTime,
            Field(description="The date and time when the member was created."),
        ]
        updated_at: Annotated[
            DateTime,
            Field(description="The date and time when the member was last updated."),
        ]



from .common import PyRevolutBaseException


class PyRevolutInvalidPayload(PyRevolutBaseException):
    """Invalid Webhook Payload -- The webhook payload is invalid."""

    def __init__(
        self,
        msg="The webhook payload is invalid. Either the payload signature is invalid or the payload timestamp is too old.",
        *args,
        **kwargs,
    ):
        super().__init__(msg, *args, **kwargs)



"""Create a payment draft to request an approval for a payment from a
business owner or admin before the payment is executed.
The business owner or admin must manually approve it in the
Revolut Business User Interface.

You can also retrieve one or all payment drafts, and delete a payment draft.

Reference link: https://developer.revolut.com/docs/business/payment-drafts
"""

# flake8: noqa: F401
from .endpoint import EndpointPaymentDraftsSync, EndpointPaymentDraftsAsync



"""This module holds all GET webhooks endpoints."""

# flake8: noqa: F401
from .retrieve_list_of_failed_webhooks import RetrieveListOfFailedWebhooks
from .retrieve_list_of_webhooks import RetrieveListOfWebhooks
from .retrieve_webhook import RetrieveWebhook



from typing import Annotated
from uuid import UUID

from pydantic import BaseModel, Field

from pyrevolut.api.common import EnumPayoutLinkState


class ResourcePayoutLinkStateChanged(BaseModel):
    """
    Payout Link State Changed resource model.
    """

    id: Annotated[
        UUID,
        Field(description="The ID of the payout link."),
    ]
    request_id: Annotated[
        str | None,
        Field(description="The request ID provided by the client."),
    ] = None
    old_state: Annotated[
        EnumPayoutLinkState,
        Field(description="The previous state of the payout link."),
    ]
    new_state: Annotated[
        EnumPayoutLinkState,
        Field(description="The new state of the payout link."),
    ]



from pydantic import BaseModel

from pyrevolut.api.webhooks.resources import ResourceWebhook


class RetrieveListOfWebhooks:
    """
    Get the list of all your existing webhooks and their details.
    """

    ROUTE = "/2.0/webhooks"

    class Params(BaseModel):
        """
        The query parameters for the request.
        """

        pass

    class Response(ResourceWebhook):
        """
        The response model for the request.
        """

        pass



"""Manage cards for the business team members, freeze, unfreeze,
terminate and update card settings, such as transaction limits.

This feature is available in the UK, US and the EEA.
This feature is not available in Sandbox.

Reference link: https://developer.revolut.com/docs/business/cards
"""

# flake8: noqa: F401
from .endpoint import EndpointCardsSync, EndpointCardsAsync



from typing import Type

from pydantic import BaseModel

from httpx import AsyncClient as HTTPClient

from pyrevolut.api import (
    EndpointAccountsAsync,
    EndpointCardsAsync,
    EndpointCounterpartiesAsync,
    EndpointForeignExchangeAsync,
    EndpointPaymentDraftsAsync,
    EndpointPayoutLinksAsync,
    EndpointSimulationsAsync,
    EndpointTeamMembersAsync,
    EndpointTransactionsAsync,
    EndpointTransfersAsync,
    EndpointWebhooksAsync,
)


from .base import BaseClient


class AsyncClient(BaseClient):
    """The asynchronous client for the Revolut API"""

    Accounts: EndpointAccountsAsync
    Cards: EndpointCardsAsync
    Counterparties: EndpointCounterpartiesAsync
    ForeignExchange: EndpointForeignExchangeAsync
    PaymentDrafts: EndpointPaymentDraftsAsync
    PayoutLinks: EndpointPayoutLinksAsync
    Simulations: EndpointSimulationsAsync
    TeamMembers: EndpointTeamMembersAsync
    Transactions: EndpointTransactionsAsync
    Transfers: EndpointTransfersAsync
    Webhooks: EndpointWebhooksAsync

    async def open(self):
        """Opens the client connection"""
        if self.client is not None:
            return

        self.client = HTTPClient()

    async def close(self):
        """Closes the client connection"""
        if self.client is None:
            return

        await self.client.aclose()
        self.client = None

    async def get(
        self,
        path: str,
        response_model: Type[BaseModel],
        params: Type[BaseModel] | None = None,
        **kwargs,
    ):
        """Send an async GET request to the Revolut API

        Parameters
        ----------
        path : str
            The path to send the request to
        response_model : Type[BaseModel]
            The model to use for the response
        params : Type[BaseModel] | None
            The parameters to send in the request

        Returns
        -------
        Response
            The response from the request
        """
        resp = await self.client.get(
            **self._prep_get(
                path=path,
                params=params,
                **kwargs,
            )
        )
        return self.process_response(
            response=resp,
            response_model=response_model,
            return_type=None,
            error_response=None,
        )

    async def post(
        self,
        path: str,
        response_model: Type[BaseModel],
        body: Type[BaseModel] | None = None,
        **kwargs,
    ):
        """Send an async POST request to the Revolut API

        Parameters
        ----------
        path : str
            The path to send the request to
        response_model : Type[BaseModel]
            The model to use for the response
        body : Type[BaseModel] | None
            The body to send in the request

        Returns
        -------
        Response
            The response from the request
        """
        resp = await self.client.post(
            **self._prep_post(
                path=path,
                body=body,
                **kwargs,
            )
        )
        return self.process_response(
            response=resp,
            response_model=response_model,
            return_type=None,
            error_response=None,
        )

    async def patch(
        self,
        path: str,
        response_model: Type[BaseModel],
        body: Type[BaseModel] | None = None,
        **kwargs,
    ):
        """Send an async PATCH request to the Revolut API

        Parameters
        ----------
        path : str
            The path to send the request to
        response_model : Type[BaseModel]
            The model to use for the response
        body : Type[BaseModel] | None
            The body to send in the request

        Returns
        -------
        Response
            The response from the request
        """
        resp = await self.client.patch(
            **self._prep_patch(
                path=path,
                body=body,
                **kwargs,
            )
        )
        return self.process_response(
            response=resp,
            response_model=response_model,
            return_type=None,
            error_response=None,
        )

    async def delete(
        self,
        path: str,
        response_model: Type[BaseModel],
        params: Type[BaseModel] | None = None,
        **kwargs,
    ):
        """Send an async DELETE request to the Revolut API

        Parameters
        ----------
        path : str
            The path to send the request to
        response_model : Type[BaseModel]
            The model to use for the response
        params : Type[BaseModel] | None
            The parameters to add to the request route

        Returns
        -------
        Response
            The response from the request
        """
        resp = await self.client.delete(
            **self._prep_delete(
                path=path,
                params=params,
                **kwargs,
            )
        )
        return self.process_response(
            response=resp,
            response_model=response_model,
            return_type=None,
            error_response=None,
        )

    async def put(
        self,
        path: str,
        response_model: Type[BaseModel],
        body: Type[BaseModel] | None = None,
        **kwargs,
    ):
        """Send an async PUT request to the Revolut API

        Parameters
        ----------
        path : str
            The path to send the request to
        response_model : Type[BaseModel]
            The model to use for the response
        body : Type[BaseModel] | None
            The body to send in the request

        Returns
        -------
        Response
            The response from the request
        """
        resp = await self.client.put(
            **self._prep_put(
                path=path,
                body=body,
                **kwargs,
            )
        )
        return self.process_response(
            response=resp,
            response_model=response_model,
            return_type=None,
            error_response=None,
        )

    def load_endpoints(self):
        """Loads all the endpoints from the api directory"""
        self.Accounts = EndpointAccountsAsync(client=self)
        self.Cards = EndpointCardsAsync(client=self)
        self.Counterparties = EndpointCounterpartiesAsync(client=self)
        self.ForeignExchange = EndpointForeignExchangeAsync(client=self)
        self.PaymentDrafts = EndpointPaymentDraftsAsync(client=self)
        self.PayoutLinks = EndpointPayoutLinksAsync(client=self)
        self.Simulations = EndpointSimulationsAsync(client=self)
        self.TeamMembers = EndpointTeamMembersAsync(client=self)
        self.Transactions = EndpointTransactionsAsync(client=self)
        self.Transfers = EndpointTransfersAsync(client=self)
        self.Webhooks = EndpointWebhooksAsync(client=self)

    async def __aenter__(self):
        """Open the async client connection"""
        await self.open()
        return self

    async def __aexit__(self, *args, **kwargs):
        """Close the async client connection"""
        await self.close()



from typing import Annotated

from pydantic import BaseModel, Field
from pydantic_extra_types.country import CountryAlpha2
from pydantic_extra_types.currency_code import Currency

from pyrevolut.api.common import EnumTransferReasonCode


class GetTransferReasons:
    """
    In order to initiate a transfer in certain currencies and countries,
    you must provide a transfer reason.
    With this endpoint you can retrieve all transfer reasons available to your business account
    per country and currency.

    After you retrieve the results, use the appropriate reason code in the transfer_reason_code
    field when making a transfer to a counterparty or creating a payout link.
    """

    ROUTE = "/1.0/transfer-reasons"

    class Params(BaseModel):
        """
        The query parameters for the request.
        """

        pass

    class Response(BaseModel):
        """
        The response model for the request.
        """

        country: Annotated[
            CountryAlpha2,
            Field(
                description="""
                The bank country of the counterparty as the 2-letter ISO 3166 code.
                """,
            ),
        ]
        currency: Annotated[
            Currency,
            Field(
                description="""
                ISO 4217 currency code in upper case.
                """,
            ),
        ]
        code: Annotated[
            EnumTransferReasonCode,
            Field(
                description="""
                Category name of the transfer reason.
                """,
            ),
        ]
        description: Annotated[
            str,
            Field(
                description="""
                The description of the given transfer reason.
                """,
            ),
        ]



from .common import PyRevolutBaseException


class PyRevolutTimeoutError(PyRevolutBaseException):
    """Timeout Error -- The request timed out."""

    def __init__(
        self,
        msg="The request timed out. Please try again.",
        *args,
        **kwargs,
    ):
        super().__init__(msg, *args, **kwargs)



from typing import Annotated

from pydantic import BaseModel, Field, model_validator

from pyrevolut.api.common import (
    EnumAccountNameMatchCode,
    EnumAccountNameMatchReasonType,
    EnumAccountNameMatchReasonCode,
)


class ValidateAccountName:
    """
    Use Confirmation of Payee (CoP) to validate a UK counterparty's account name
    against their account number and sort code when adding a counterparty or making a
    transfer to a new or existing counterparty.

    Note
    ----
    Confirmation of Payee is an account name checking system in the UK that helps clients
    to make sure payments aren't sent to the wrong bank or building society account.

    When performing the check, you must specify the account type by providing the name for either
    an individual (personal account) or a company (business account).

    Caution
    -------
    The CoP check does not protect you against all kinds of fraud. It only checks if the name you provided for an account matches that account's details.
    Even if the counterparty's details match, you should still exercise due caution when transferring funds.

    This functionality is only available to UK-based businesses.
    """

    ROUTE = "/1.0/account-name-validation"

    class Body(BaseModel):
        """
        Request body for the endpoint.
        """

        class ModelIndividualName(BaseModel):
            """The name of the individual counterparty. Use when company_name isn't specified."""

            first_name: Annotated[
                str | None,
                Field(description="The first name of the individual counterparty."),
            ] = None
            last_name: Annotated[
                str | None,
                Field(description="The last name of the individual counterparty."),
            ] = None

        account_no: Annotated[
            str,
            Field(
                description="The account number of the counterparty.",
            ),
        ]
        sort_code: Annotated[
            str,
            Field(
                description="The sort code of the counterparty's account.",
            ),
        ]
        company_name: Annotated[
            str | None,
            Field(
                description="The name of the business counterparty. Use when individual_name is not specified.",
            ),
        ] = None
        individual_name: Annotated[
            ModelIndividualName | None,
            Field(
                description="The name of the individual counterparty. Use when company_name is not specified.",
            ),
        ] = None

        @model_validator(mode="after")
        def check_inputs(self) -> "ValidateAccountName.Body":
            """
            Ensure that either the individual_name or company_name is provided.
            """
            if not self.company_name and not self.individual_name:
                raise ValueError(
                    "You must provide either the company_name or individual_name."
                )
            if self.company_name and self.individual_name:
                raise ValueError(
                    "You must provide either the company_name or individual_name, not both."
                )
            return self

    class Response(BaseModel):
        """
        Response model for the endpoint.
        """

        class ModelIndividualName(BaseModel):
            """The name of the individual counterparty. Use when company_name isn't specified."""

            first_name: Annotated[
                str | None,
                Field(description="The first name of the individual counterparty."),
            ] = None
            last_name: Annotated[
                str | None,
                Field(description="The last name of the individual counterparty."),
            ] = None

        class ModelReason(BaseModel):
            """
            A code which explains why a given result was returned.
            For example, it might happen that the details you provided match the account details,
            but you specified the counterparty as an individual, and the account type is business.
            """

            type: Annotated[
                EnumAccountNameMatchReasonType | None,
                Field(
                    description="""
                    The reason type. Possible values:

                        uk_cop:
                            The CoP reason
                    """,
                ),
            ] = None
            code: Annotated[
                EnumAccountNameMatchReasonCode | None,
                Field(
                    description="""
                    The reason code. Possible values:

                        close_match:
                            The provided name is similar to the account name, the account type is correct.
                            The actual name is returned.
                        individual_account_name_matched:
                            The names match but the counterparty is an individual, not a business.
                        company_account_name_matched:
                            The names match but the counterparty is a business, not an individual.
                        individual_account_close_match:
                            The provided name is similar to the account name, and the account type is incorrect
                            - the counterparty is an individual, not a business. The actual name is returned.
                        company_account_close_match:
                            The provided name is similar to the account name, and the account type is incorrect
                            - the counterparty is a business, not an individual. The actual name is returned.
                        not_matched:
                            The account details don't match the provided values.
                        account_does_not_exist:
                            The account does not exist.
                        account_switched:
                            The account has been switched using the Current Account Switching Service.
                            Please contact the recipient for updated account details.
                        cannot_be_checked:
                            The account cannot be checked.
                    """,
                ),
            ] = None

        result_code: Annotated[
            EnumAccountNameMatchCode,
            Field(
                description="""
                The result of the account name check. Possible values:

                    matched:
                        The name and account type match the provided values.
                    close_match:
                        The name and account type are similar to the provided values.
                        The actual values are returned.
                    not_matched:
                        The name and account type don't match the provided values.
                    cannot_be_checked:
                        The check cannot be performed and retries won't help.
                        For example, the recipient's bank doesn't support CoP.
                    temporarily_unavailable:
                        The check cannot be performed right now.
                        For example, the recipient's bank didn't respond to our request.
                        You should retry the request later.
                """
            ),
        ]
        reason: Annotated[
            ModelReason | None,
            Field(
                description="""
                A code which explains why a given result was returned.
                For example, it might happen that the details you provided match the account details,
                but you specified the counterparty as an individual, and the account type is business.
                """,
            ),
        ] = None
        company_name: Annotated[
            str | None,
            Field(
                description="The name of the business counterparty. Use when individual_name is not specified.",
            ),
        ] = None
        individual_name: Annotated[
            ModelIndividualName | None,
            Field(
                description="The name of the individual counterparty. Use when company_name is not specified.",
            ),
        ] = None



"""A webhook (also called a web callback) allows your system to receive
updates about your account to an HTTPS endpoint that you provide.
When a supported event occurs, a notification is posted via HTTP POST method
to the specified endpoint.

If the receiver returns an HTTP error response, Revolut will retry the webhook
event three more times, each with a 10-minute interval.

The following events are supported:

TransactionCreated
TransactionStateChanged
PayoutLinkCreated
PayoutLinkStateChanged

Reference link: https://developer.revolut.com/docs/business/webhooks-v-2
"""

# flake8: noqa: F401
from .endpoint import EndpointWebhooksSync, EndpointWebhooksAsync



from typing import Annotated
from uuid import UUID

from pydantic import BaseModel, Field, ConfigDict

from pyrevolut.utils import DateTime, Date
from pyrevolut.api.common import EnumTransactionType
from pyrevolut.api.transactions.resources import ResourceTransaction


class RetrieveListOfTransactions:
    """
    Retrieve the historical transactions based on the provided query criteria.

    The transactions are sorted by the created_at date in reverse chronological order,
    and they're paginated. The maximum number of transactions returned per page is specified by the
    count parameter. To get the next page of results, make a new request and use the created_at date
    from the last item of the previous page as the value for the to parameter.

    Note
    ----
    The API returns a maximum of 1,000 transactions per request.

    Note
    ----
    To be compliant with PSD2 SCA regulations, businesses on the Revolut Business Freelancer
    plans can only access information older than 90 days within 5 minutes of the first authorisation.
    """

    ROUTE = "/1.0/transactions"

    class Params(BaseModel):
        """
        The query parameters for the request.
        """

        model_config = ConfigDict(
            populate_by_name=True,
            from_attributes=True,
        )

        from_: Annotated[
            DateTime | Date | None,
            Field(
                description="""
                The date and time you retrieve the historical transactions from, including 
                this date-time. 
                Corresponds to the created_at value of the transaction. 
                Provided in ISO 8601 format.

                Used also for pagination. To get back to the previous page of results, 
                make a new request and use the created_at date from the first item of the 
                current page as the value for the from parameter.
                """
            ),
        ] = None
        to: Annotated[
            DateTime | Date | None,
            Field(
                description="""
                The date and time you retrieve the historical transactions to, excluding 
                this date-time. 
                Corresponds to the created_at value of the transaction. 
                Provided in ISO 8601 format. 
                The default value is the date and time at which you're calling the endpoint.

                Used also for pagination. 
                To get the next page of results, make a new request and use the created_at 
                date from the last item of the previous (current) page as the value for the 
                to parameter.  
                """
            ),
        ] = None
        account: Annotated[
            UUID | None,
            Field(
                description="""
                The ID of the account
                """
            ),
        ] = None
        count: Annotated[
            int | None,
            Field(
                description="""
                The maximum number of the historical transactions to retrieve per page.

                To get the next page of results, make a new request and use 
                the created_at date from the last item of the previous page as the 
                value for the to parameter.   
                
                If not provided, the default value is 100.             
                """,
                le=1000,
                ge=1,
            ),
        ] = None
        type: Annotated[
            EnumTransactionType | None,
            Field(
                description="""
                The type of the historical transactions to retrieve.
                """
            ),
        ] = None

    class Response(ResourceTransaction):
        """
        The response model for the request.
        """

        pass



from typing import Annotated
from uuid import UUID

from pydantic import BaseModel, Field

from pyrevolut.utils import Date


class RetrieveAllPaymentDrafts:
    """
    Get a list of all the payment drafts that aren't processed.
    """

    ROUTE = "/1.0/payment-drafts"

    class Params(BaseModel):
        """
        The parameters for the endpoint.
        """

        pass

    class Response(BaseModel):
        """
        The response model for the endpoint.
        """

        class ModelPaymentOrder(BaseModel):
            """A payment draft."""

            id: Annotated[UUID, Field(description="The ID of the payment draft.")]
            scheduled_for: Annotated[
                Date | None,
                Field(
                    description="The scheduled date of the payment draft in ISO 8601 format."
                ),
            ] = None
            title: Annotated[
                str | None,
                Field(description="The title of the payment draft."),
            ] = None
            payment_counts: Annotated[
                int,
                Field(description="The number of payments in the payment draft.", ge=0),
            ]

        payment_orders: Annotated[
            list[ModelPaymentOrder],
            Field(description="The list of payment drafts."),
        ]



from typing import Annotated
from uuid import UUID

from pydantic import BaseModel, Field, model_validator

from pyrevolut.api.common import ModelBaseAmount, EnumMerchantCategory
from pyrevolut.api.cards.resources import ResourceCard


class CreateCard:
    """
    Create a new card for an existing member of your Revolut Business team.

    When using the API, you can create only virtual cards.
    To create a physical card, use the Revolut Business app.
    """

    ROUTE = "/1.0/cards"

    class Body(BaseModel):
        """
        Body model for the endpoint.
        """

        class ModelSpendingLimits(BaseModel):
            """All spending limits set for the card."""

            class ModelSingle(ModelBaseAmount):
                """The limit for a single transaction."""

                pass

            class ModelDay(ModelBaseAmount):
                """The daily limit for transactions."""

                pass

            class ModelWeek(ModelBaseAmount):
                """The weekly limit for transactions."""

                pass

            class ModelMonth(ModelBaseAmount):
                """The monthly limit for transactions."""

                pass

            class ModelQuarter(ModelBaseAmount):
                """The quarterly limit for transactions."""

                pass

            class ModelYear(ModelBaseAmount):
                """The yearly limit for transactions."""

                pass

            class ModelAllTime(ModelBaseAmount):
                """The all-time limit for transactions."""

                pass

            single: Annotated[
                ModelSingle | None,
                Field(description="The limit for a single transaction."),
            ]
            day: Annotated[
                ModelDay | None,
                Field(description="The daily limit for transactions."),
            ]
            week: Annotated[
                ModelWeek | None,
                Field(description="The weekly limit for transactions."),
            ]
            month: Annotated[
                ModelMonth | None,
                Field(description="The monthly limit for transactions."),
            ]
            quarter: Annotated[
                ModelQuarter | None,
                Field(description="The quarterly limit for transactions."),
            ]
            year: Annotated[
                ModelYear | None,
                Field(description="The yearly limit for transactions."),
            ]
            all_time: Annotated[
                ModelAllTime | None,
                Field(description="The all-time limit for transactions."),
            ]

        request_id: Annotated[
            str,
            Field(
                description="""
                A unique ID of the request that you provide.
                This ID is used to prevent duplicate card creation requests in case 
                of a lost connection or client error, so make sure you use the same 
                request_id for requests related to the same card.
                The deduplication is limited to 24 hours counting from the first request 
                using a given ID. 
                """,
                max_length=40,
            ),
        ]
        virtual: Annotated[
            bool,
            Field(
                description="""
                Specifies the type of the card. Must be set to true, as with the API, you can create only virtual cards.
                """,
            ),
        ] = True
        holder_id: Annotated[
            UUID,
            Field(
                description="""
                The ID of the team member who will be the holder of the card.
                """
            ),
        ]
        label: Annotated[
            str | None,
            Field(
                description="""
                The label for the issued card, displayed in the UI to help distinguish between cards. 
                If not specified, no label will be added.
                """,
                max_length=30,
            ),
        ] = None
        accounts: Annotated[
            list[UUID] | None,
            Field(
                description="""
                The list of accounts to link to the card. If not specified, all accounts will be linked.
                """,
            ),
        ] = None
        categories: Annotated[
            list[EnumMerchantCategory] | None,
            Field(
                description="""
                The list of merchant categories to be available for card spending. 
                If not specified, all categories will be allowed.
                """,
            ),
        ] = None
        spending_limits: Annotated[
            ModelSpendingLimits | None,
            Field(
                description="""
                All spending limits set for the card.
                You can have at most 1 periodic (day/week/month/quarter/all-time) and 
                1 non-periodic (single transaction) limit at a time. 
                If you try to specify 2 periodic limits at a time, it will result in an error.

                Spending limit currency must match the default business currency. 
                The default currency was assigned to your business during onboarding.
                """,
            ),
        ] = None

        @model_validator(mode="after")
        def check_inputs(self) -> "CreateCard.Body":
            """Check the inputs."""
            if not self.virtual:
                raise ValueError("You can only create virtual cards via the API.")
            if self.spending_limits is not None:
                if (
                    sum(
                        [
                            self.spending_limits.day is not None,
                            self.spending_limits.week is not None,
                            self.spending_limits.month is not None,
                            self.spending_limits.quarter is not None,
                            self.spending_limits.year is not None,
                            self.spending_limits.all_time is not None,
                        ]
                    )
                    > 1
                ):
                    raise ValueError(
                        "You can have at most 1 periodic (day/week/month/quarter/all-time) limit at a time."
                    )
            return self

    class Response(ResourceCard):
        """
        Response model for the endpoint.
        """

        pass



"""Get the balances, full banking details, and other details of your business accounts.
Reference link: https://developer.revolut.com/docs/business/accounts
"""

# flake8: noqa: F401
from .endpoint import EndpointAccountsSync, EndpointAccountsAsync



from uuid import UUID
from datetime import datetime

from pyrevolut.api.common import BaseEndpointSync, EnumProfileType
from pyrevolut.utils import DateTime

from pyrevolut.api.counterparties.get import (
    RetrieveListOfCounterparties,
    RetrieveCounterparty,
)
from pyrevolut.api.counterparties.post import CreateCounterparty, ValidateAccountName
from pyrevolut.api.counterparties.delete import DeleteCounterparty


class EndpointCounterpartiesSync(BaseEndpointSync):
    """The Counterparties API

    Manage counterparties that you intend to transact with.

    Request and response examples can vary based on the account provider's
    location and type of the counterparty.

    In the Sandbox environment, you cannot add real people and businesses as Revolut counterparties.
    Therefore, to help you simulate Create a counterparty requests, we have created some
    test accounts for counterparties of profile type personal.

    To add a counterparty via Revtag, use one of these pairs for the name and revtag fields respectively:

    Test User 1 & john1pvki
    Test User 2 & john2pvki
    ...
    Test User 9 & john9pvki
    """

    def get_all_counterparties(
        self,
        name: str | None = None,
        account_no: str | None = None,
        sort_code: str | None = None,
        iban: str | None = None,
        bic: str | None = None,
        created_before: datetime | DateTime | str | int | float | None = None,
        limit: int | None = None,
        **kwargs,
    ) -> list[dict] | list[RetrieveListOfCounterparties.Response]:
        """
        Get all the counterparties that you have created, or use the query parameters to filter the results.

        The counterparties are sorted by the created_at date in reverse chronological order.

        The returned counterparties are paginated. The maximum number of counterparties returned per page
        is specified by the limit parameter. To get to the next page, make a new request and use the
        created_at date of the last counterparty returned in the previous response.

        Parameters
        ----------
        name : str | None
            The name of the counterparty to retrieve. It does not need to be an exact match,
            partial match is also supported.
        account_no : str | None
            The exact account number of the counterparty to retrieve.
        sort_code : str | None
            The exact sort code of the counterparty to retrieve.
            Only allowed in combination with the account_no parameter.
        iban : str | None
            The exact IBAN of the counterparty to retrieve.
        bic : str | None
            The exact BIC of the counterparty to retrieve. Only allowed in combination with the iban parameter.
        created_before : datetime | DateTime | str | int | float | None
            Retrieves counterparties with created_at < created_before.
            The default value is the current date and time at which you are calling the endpoint.
            Provided in ISO 8601 format.
        limit : int | None
            The maximum number of counterparties returned per page.
            To get to the next page, make a new request and use the
            created_at date of the last card returned in the previous
            response as the value for created_before.

        Returns
        -------
        list[dict] | list[RetrieveListOfCounterparties.Response]
            The list of all counterparties that you have created.
        """
        endpoint = RetrieveListOfCounterparties
        path = endpoint.ROUTE
        params = endpoint.Params(
            name=name,
            account_no=account_no,
            sort_code=sort_code,
            iban=iban,
            bic=bic,
            created_before=created_before,
            limit=limit,
        )

        return self.client.get(
            path=path,
            response_model=endpoint.Response,
            params=params,
            **kwargs,
        )

    def get_counterparty(
        self,
        counterparty_id: UUID,
        **kwargs,
    ) -> dict | RetrieveCounterparty.Response:
        """Get the information about a specific counterparty by ID.

        Parameters
        ----------
        counterparty_id : UUID
            The ID of the counterparty to retrieve.

        Returns
        -------
        dict | RetrieveCounterparty.Response
            The information about the counterparty.
        """
        endpoint = RetrieveCounterparty
        path = endpoint.ROUTE.format(counterparty_id=counterparty_id)
        params = endpoint.Params()

        return self.client.get(
            path=path,
            response_model=endpoint.Response,
            params=params,
            **kwargs,
        )

    def create_counterparty(
        self,
        company_name: str | None = None,
        profile_type: EnumProfileType | None = None,
        name: str | None = None,
        individual_first_name: str | None = None,
        individual_last_name: str | None = None,
        bank_country: str | None = None,
        currency: str | None = None,
        revtag: str | None = None,
        account_no: str | None = None,
        iban: str | None = None,
        sort_code: str | None = None,
        routing_number: str | None = None,
        bic: str | None = None,
        clabe: str | None = None,
        isfc: str | None = None,
        bsb_code: str | None = None,
        address_street_line1: str | None = None,
        address_street_line2: str | None = None,
        address_region: str | None = None,
        address_city: str | None = None,
        address_country: str | None = None,
        address_postcode: str | None = None,
        **kwargs,
    ) -> dict | CreateCounterparty.Response:
        """
        Create a new counterparty to transact with.

        In the Sandbox environment, you cannot add real people and businesses as Revolut counterparties.
        To help you simulate Create a counterparty requests for counterparties of profile type personal,
        we created some test accounts. Look inside for test Revtags.

        To add a counterparty via Revtag, use one of these pairs for the name and revtag fields respectively:

            Test User 1 & john1pvki
            Test User 2 & john2pvki
            ...
            Test User 9 & john9pvki

        Parameters
        ----------
        company_name : str | None
            The name of the company counterparty.
            Use when individual_name or name isn't specified and profile_type is business.
        profile_type : EnumProfileType | None
            The type of the Revolut profile. Used when adding an existing Revolut user via Revtag.
        name : str | None
            The name of the counterparty that you create for an existing Revolut user via Revtag.
            Provide the value only when you specify personal for profile_type.
        individual_first_name : str | None
            The first name of the individual counterparty.
            Use when company_name isn't specified.
        individual_last_name : str | None
            The last name of the individual counterparty.
            Use when company_name isn't specified.
        bank_country : str | None
            The country of the counterparty's bank as the 2-letter ISO 3166 code.
        currency : str | None
            ISO 4217 currency code in upper case.
        revtag : str | None
            The Revtag of the counterparty to add.
        account_no : str | None
            The bank account number of the counterparty.
        iban : str | None
            The IBAN number of the counterparty's account. This field is displayed for IBAN countries.
        sort_code : str | None
            The sort code of the counterparty's bank. This field is displayed for GBP accounts.
        routing_number : str | None
            The routing number of the counterparty's bank. This field is displayed for USD accounts.
        bic : str | None
            The BIC number of the counterparty's account. This field is required for non-SEPA IBAN/SWIFT.
        clabe : str | None
            The CLABE number of the counterparty's account. This field is required for SWIFT MX.
        isfc : str | None
            The ISFC number of the counterparty's account. This field is required for INR accounts.
        bsb_code : str | None
            The BSB code of the counterparty's account. This field is required for AUD accounts.
        address_street_line1 : str | None
            Street line 1 information.
        address_street_line2 : str | None
            Street line 2 information.
        address_region : str | None
            The name of the region.
        address_city : str | None
            The name of the city.
        address_country : str | None
            The country of the counterparty's address as the 2-letter ISO 3166 code.
        address_postcode : str | None
            The postcode of the counterparty's address.

        Returns
        -------
        dict | CreateCounterparty.Response
            A dict with the information about the created counterparty.
        """
        endpoint = CreateCounterparty
        path = endpoint.ROUTE
        body = endpoint.Body(
            company_name=company_name,
            profile_type=profile_type,
            name=name,
            individual_name=(
                endpoint.Body.ModelIndividualName(
                    first_name=individual_first_name,
                    last_name=individual_last_name,
                )
                if individual_first_name is not None or individual_last_name is not None
                else None
            ),
            bank_country=bank_country,
            currency=currency,
            revtag=revtag,
            account_no=account_no,
            iban=iban,
            sort_code=sort_code,
            routing_number=routing_number,
            bic=bic,
            clabe=clabe,
            isfc=isfc,
            bsb_code=bsb_code,
            address=(
                endpoint.Body.ModelAddress(
                    street_line1=address_street_line1,
                    street_line2=address_street_line2,
                    region=address_region,
                    city=address_city,
                    country=address_country,
                    postcode=address_postcode,
                )
                if address_country is not None and address_postcode is not None
                else None
            ),
        )

        return self.client.post(
            path=path,
            response_model=endpoint.Response,
            body=body,
            **kwargs,
        )

    def validate_account_name(
        self,
        account_no: str,
        sort_code: str,
        company_name: str | None = None,
        individual_first_name: str | None = None,
        individual_last_name: str | None = None,
        **kwargs,
    ) -> dict | ValidateAccountName.Response:
        """
        Use Confirmation of Payee (CoP) to validate a UK counterparty's account name
        against their account number and sort code when adding a counterparty or making a
        transfer to a new or existing counterparty.

        Note
        ----
        Confirmation of Payee is an account name checking system in the UK that helps clients
        to make sure payments aren't sent to the wrong bank or building society account.

        When performing the check, you must specify the account type by providing the name for either
        an individual (personal account) or a company (business account).

        Caution
        -------
        The CoP check does not protect you against all kinds of fraud. It only checks if the name you provided for an account matches that account's details.
        Even if the counterparty's details match, you should still exercise due caution when transferring funds.

        This functionality is only available to UK-based businesses.

        Parameters
        ----------
        account_no : str
            The account number of the counterparty.
        sort_code : str
            The sort code of the counterparty's account.
        company_name : str | None
            The name of the business counterparty. Use when individual_name is not specified.
        individual_first_name : str | None
            The first name of the individual counterparty.
            Use when company_name isn't specified.
        individual_last_name : str | None
            The last name of the individual counterparty.
            Use when company_name isn't specified.

        Returns
        -------
        dict | ValidateAccountName.Response
            A dict with the information about the validated account name.
        """
        endpoint = ValidateAccountName
        path = endpoint.ROUTE
        body = endpoint.Body(
            account_no=account_no,
            sort_code=sort_code,
            company_name=company_name,
            individual_name=(
                endpoint.Body.ModelIndividualName(
                    first_name=individual_first_name,
                    last_name=individual_last_name,
                )
                if individual_first_name is not None or individual_last_name is not None
                else None
            ),
        )

        return self.client.post(
            path=path,
            response_model=endpoint.Response,
            body=body,
            **kwargs,
        )

    def delete_counterparty(
        self,
        counterparty_id: UUID,
        **kwargs,
    ) -> dict | DeleteCounterparty.Response:
        """Delete a counterparty with the given ID.
        When a counterparty is deleted, you cannot make any payments to the counterparty.

        Parameters
        ----------
        counterparty_id : UUID
            The ID of the counterparty to delete.

        Returns
        -------
        dict | DeleteCounterparty.Response
            An empty dict.
        """
        endpoint = DeleteCounterparty
        path = endpoint.ROUTE.format(counterparty_id=counterparty_id)
        params = endpoint.Params()

        return self.client.delete(
            path=path,
            response_model=endpoint.Response,
            params=params,
            **kwargs,
        )



from pydantic import BaseModel

from pyrevolut.api.accounts.resources import ResourceAccount


class RetrieveAllAccounts:
    """
    Get a list of all your accounts.
    """

    ROUTE = "/1.0/accounts"

    class Params(BaseModel):
        """
        Query parameters for the endpoint.
        """

        pass

    class Response(ResourceAccount):
        """
        Response model for the endpoint.
        """

        pass



from typing import Annotated
from uuid import UUID

from pydantic import BaseModel, Field
from pydantic_extra_types.currency_code import Currency

from pyrevolut.api.common import EnumChargeBearer, EnumTransferReasonCode
from pyrevolut.api.transfers.resources import ResourceTransfer


class CreateTransferToAnotherAccount:
    """
    Make a payment to a counterparty.
    The resulting transaction has the type transfer.

    If you make the payment to another Revolut account, either business or personal,
    the transaction is executed instantly.

    If the counterparty has multiple payment methods available, for example, 2 accounts,
    or 1 account and 1 card, you must specify the account or card to which you want to
    transfer the money (receiver.account_id or receiver.card_id respectively).

    Caution
    -------
    Due to PSD2 Strong Customer Authentication regulations, this endpoint is
    only available for customers on Revolut Business Company plans. If you're a
    freelancer and wish to make payments via our API, we advise that you instead
    leverage our Payment drafts (/payment-drafts) endpoint.
    """

    ROUTE = "/1.0/pay"

    class Body(BaseModel):
        """
        The body of the request.
        """

        class ModelReceiver(BaseModel):
            """
            The details of the transfer recipient.

            If the counterparty has multiple payment methods available
            (e.g. 2 accounts, or 1 account and 1 card), you must specify the account
            (account_id) or card (card_id) to which you want to transfer the money.
            """

            counterparty_id: Annotated[
                UUID,
                Field(description="The ID of the receiving counterparty."),
            ]
            account_id: Annotated[
                UUID | None,
                Field(
                    description="""
                    The ID of the receiving counterparty's account, which can be own account. 
                    Used for bank transfers.

                    If the counterparty has multiple payment methods available, use it to 
                    specify the account to which you want to send the money.
                    """,
                ),
            ] = None
            card_id: Annotated[
                UUID | None,
                Field(
                    description="""
                    The ID of the receiving counterparty's card. Used for card transfers.

                    If the counterparty has multiple payment methods available, use it to 
                    specify the card to which you want to send the money.
                    """,
                ),
            ] = None

        request_id: Annotated[
            str,
            Field(
                description="""
                The ID of the request, provided by you. 
                It helps you identify the transaction in your system.
                To ensure that a transfer is not processed multiple times if 
                there are network or system errors, the same request_id should be used 
                for requests related to the same transfer.
                """,
                max_length=40,
            ),
        ]
        account_id: Annotated[
            UUID,
            Field(description="The ID of the account that you send the funds from."),
        ]
        receiver: Annotated[
            ModelReceiver,
            Field(
                description="""
                The details of the transfer recipient.

                If the counterparty has multiple payment methods available
                (e.g. 2 accounts, or 1 account and 1 card), you must specify the account
                (account_id) or card (card_id) to which you want to transfer the money.
                """
            ),
        ]
        amount: Annotated[
            float,
            Field(
                description="The amount of the funds to be transferred.",
                gt=0,
            ),
        ]
        currency: Annotated[
            Currency,
            Field(description="The ISO 4217 currency of the funds to be transferred."),
        ]
        reference: Annotated[
            str | None,
            Field(
                description="""
                The reference for the transaction.
                """,
            ),
        ] = None
        charge_bearer: Annotated[
            EnumChargeBearer | None,
            Field(
                description="""
                The party to which any transaction fees are charged if the resulting 
                transaction route has associated fees. Some transactions with fees might 
                not be possible with the specified option, in which case error 3287 is returned.

                Possible values:
                - shared: The transaction fees are shared between the sender and the receiver.
                - debtor: The sender pays the transaction fees.
                """,
            ),
        ] = None
        transfer_reason_code: Annotated[
            EnumTransferReasonCode | None,
            Field(
                description="""
                The reason code for the transaction. 
                Transactions to certain countries and currencies might require 
                you to provide a transfer reason. 
                You can check available reason codes with the getTransferReasons operation.

                If a transfer reason is not required for the given currency and country, 
                this field is ignored.
                """,
            ),
        ] = None

    class Response(ResourceTransfer):
        """
        The response model of the request.
        """

        pass



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyrevolut.client import Client, AsyncClient


class BaseEndpointSync:
    """Base class for all endpoints."""

    def __init__(self, client: "Client"):
        """Create a new Base endpoint handler

        Parameters
        ----------
        client : Client
            The client to use for the endpoint
        """
        self.client = client


class BaseEndpointAsync:
    """Base  class for all async endpoints."""

    def __init__(self, client: "AsyncClient"):
        """Create a new Base endpoint handler

        Parameters
        ----------
        client : AsyncClient
            The async client to use for the endpoint
        """
        self.client = client



from typing import Annotated
from uuid import UUID

from pydantic import BaseModel, Field
from pydantic_extra_types.currency_code import Currency

from pyrevolut.api.transfers.resources import ResourceTransfer


class MoveMoneyBetweenAccounts:
    """
    Move money between the Revolut accounts of the business in the same currency.

    The resulting transaction has the type transfer.
    """

    ROUTE = "/1.0/transfer"

    class Body(BaseModel):
        """
        The body of the request.
        """

        request_id: Annotated[
            str,
            Field(
                description="""
                The ID of the request, provided by you. 
                It helps you identify the transaction in your system.
                To ensure that a transfer is not processed multiple times if 
                there are network or system errors, the same request_id should be used 
                for requests related to the same transfer.
                """,
                max_length=40,
            ),
        ]
        source_account_id: Annotated[
            UUID,
            Field(
                description="The ID of the source account that you transfer the funds from."
            ),
        ]
        target_account_id: Annotated[
            UUID,
            Field(
                description="The ID of the target account that you transfer the funds to."
            ),
        ]
        amount: Annotated[
            float,
            Field(
                description="The amount of the funds to be transferred.",
                gt=0,
            ),
        ]
        currency: Annotated[
            Currency,
            Field(description="The ISO 4217 currency of the funds to be transferred."),
        ]
        reference: Annotated[
            str | None,
            Field(
                description="""
                The reference for the funds transfer.
                """,
            ),
        ] = None

    class Response(ResourceTransfer):
        """
        The response model of the request.
        """

        pass



from enum import StrEnum


class EnumPaymentDraftState(StrEnum):
    """Indicates the payment draft state."""

    CREATED = "CREATED"
    PENDING = "PENDING"
    COMPLETED = "COMPLETED"
    REVERTED = "REVERTED"
    DECLINED = "DECLINED"
    CANCELLED = "CANCELLED"
    FAILED = "FAILED"
    DELETED = "DELETED"



from typing import Annotated

from pydantic import BaseModel, Field, HttpUrl

from pyrevolut.api.common import EnumWebhookEvent
from pyrevolut.api.webhooks.resources import ResourceWebhook


class CreateWebhook:
    """
    Create a new webhook to receive event notifications to the specified URL.
    Provide a list of event types that you want to subscribe to and a URL for the webhook.
    Only HTTPS URLs are supported.
    """

    ROUTE = "/2.0/webhooks"

    class Body(BaseModel):
        """
        The body of the request.
        """

        url: Annotated[
            HttpUrl,
            Field(
                description="""
                A valid webhook URL to which to send event notifications. 
                The supported protocol is https.
                """,
            ),
        ]
        events: Annotated[
            list[EnumWebhookEvent] | None,
            Field(
                description="""
                A list of event types to subscribe to. 
                If you don't provide it, you're automatically subscribed to the default event types:
                - TransactionCreated
                - TransactionStateChanged
                """,
            ),
        ] = None

    class Response(ResourceWebhook):
        """
        The response model for the request.
        """

        signing_secret: Annotated[
            str,
            Field(
                description="""
                The signing secret for the webhook.
                """,
            ),
        ]



from typing import Annotated

from pydantic import BaseModel, Field
from pydantic_extra_types.currency_code import Currency


class ModelBaseAmount(BaseModel):
    """Base model for amount"""

    amount: Annotated[float, Field(description="The value.")]
    currency: Annotated[
        Currency,
        Field(description="The currency, provided as a 3-letter ISO 4217 code."),
    ]



"""This module holds all GET card endpoints."""

# flake8: noqa: F401
from .retrieve_card_details import RetrieveCardDetails
from .retrieve_list_of_cards import RetrieveListOfCards
from .retrieve_sensitive_card_details import RetrieveSensitiveCardDetails



from uuid import UUID

from pyrevolut.api.common import (
    BaseEndpointSync,
    EnumChargeBearer,
    EnumTransferReasonCode,
)

from pyrevolut.api.transfers.get import GetTransferReasons
from pyrevolut.api.transfers.post import (
    CreateTransferToAnotherAccount,
    MoveMoneyBetweenAccounts,
)


class EndpointTransfersSync(BaseEndpointSync):
    """The Transfers API

    Move funds in the same currency between accounts of your business,
    or make payments to your counterparties.
    """

    def get_transfer_reasons(
        self,
        **kwargs,
    ) -> list[dict] | list[GetTransferReasons.Response]:
        """
        In order to initiate a transfer in certain currencies and countries,
        you must provide a transfer reason.
        With this endpoint you can retrieve all transfer reasons available to your business account
        per country and currency.

        After you retrieve the results, use the appropriate reason code in the transfer_reason_code
        field when making a transfer to a counterparty or creating a payout link.

        Parameters
        ----------
        None

        Returns
        -------
        list[dict] | list[GetTransferReasons.Response]
            A list of transfer reasons.
        """
        endpoint = GetTransferReasons
        path = endpoint.ROUTE
        params = endpoint.Params()

        return self.client.get(
            path=path,
            response_model=endpoint.Response,
            params=params,
            **kwargs,
        )

    def create_transfer_to_another_account(
        self,
        request_id: str,
        account_id: UUID,
        counterparty_id: UUID,
        amount: float,
        currency: str,
        counterparty_account_id: UUID | None = None,
        counterparty_card_id: UUID | None = None,
        reference: str | None = None,
        charge_bearer: EnumChargeBearer | None = None,
        transfer_reason_code: EnumTransferReasonCode | None = None,
        **kwargs,
    ) -> dict | CreateTransferToAnotherAccount.Response:
        """
        Make a payment to a counterparty.
        The resulting transaction has the type transfer.

        If you make the payment to another Revolut account, either business or personal,
        the transaction is executed instantly.

        If the counterparty has multiple payment methods available, for example, 2 accounts,
        or 1 account and 1 card, you must specify the account or card to which you want to
        transfer the money (receiver.account_id or receiver.card_id respectively).

        Caution
        -------
        Due to PSD2 Strong Customer Authentication regulations, this endpoint is
        only available for customers on Revolut Business Company plans. If you're a
        freelancer and wish to make payments via our API, we advise that you instead
        leverage our Payment drafts (/payment-drafts) endpoint.

        Parameters
        ----------
        request_id : str
            The ID of the request, provided by you.
            It helps you identify the transaction in your system.
            To ensure that a transfer is not processed multiple times if
            there are network or system errors, the same request_id should be used
            for requests related to the same transfer.
        account_id : UUID
            The ID of the account that you transfer the funds from.
        counterparty_id : UUID
            The ID of the receiving counterparty.
        amount : float
            The amount of money to transfer.
        currency : str
            The currency of the transfer.
        counterparty_account_id : UUID, optional
            The ID of the receiving counterparty's account, which can be own account.
            Used for bank transfers.
            If the counterparty has multiple payment methods available, use it to
            specify the account to which you want to send the money.
        counterparty_card_id : UUID, optional
            The ID of the receiving counterparty's card. Used for card transfers.
            If the counterparty has multiple payment methods available, use it to
            specify the card to which you want to send the money.
        reference : str, optional
            A reference for the transfer.
        charge_bearer : EnumChargeBearer, optional
            The party to which any transaction fees are charged if the resulting
            transaction route has associated fees. Some transactions with fees might
            not be possible with the specified option, in which case error 3287 is returned.

            Possible values:
            - shared: The transaction fees are shared between the sender and the receiver.
            - debtor: The sender pays the transaction fees.
        transfer_reason_code : EnumTransferReasonCode, optional
            The reason code for the transaction.
            Transactions to certain countries and currencies might require
            you to provide a transfer reason.
            You can check available reason codes with the getTransferReasons operation.

            If a transfer reason is not required for the given currency and country,
            this field is ignored.

        Returns
        -------
        dict | CreateTransferToAnotherAccount.Response
            The details of the transfer.
        """
        endpoint = CreateTransferToAnotherAccount
        path = endpoint.ROUTE
        body = endpoint.Body(
            request_id=request_id,
            account_id=account_id,
            receiver=endpoint.Body.ModelReceiver(
                counterparty_id=counterparty_id,
                account_id=counterparty_account_id,
                card_id=counterparty_card_id,
            ),
            amount=amount,
            currency=currency,
            reference=reference,
            charge_bearer=charge_bearer,
            transfer_reason_code=transfer_reason_code,
        )

        return self.client.post(
            path=path,
            response_model=endpoint.Response,
            body=body,
            **kwargs,
        )

    def move_money_between_accounts(
        self,
        request_id: str,
        source_account_id: UUID,
        target_account_id: UUID,
        amount: float,
        currency: str,
        reference: str | None = None,
        **kwargs,
    ) -> dict | MoveMoneyBetweenAccounts.Response:
        """
        Move money between the Revolut accounts of the business in the same currency.

        The resulting transaction has the type transfer.

        Parameters
        ----------
        request_id : str
            The ID of the request, provided by you.
            It helps you identify the transaction in your system.
            To ensure that a transfer is not processed multiple times if
            there are network or system errors, the same request_id should be used
            for requests related to the same transfer.
        source_account_id : UUID
            The ID of the source account that you transfer the funds from.
        target_account_id : UUID
            The ID of the target account that you transfer the funds to.
        amount : float
            The amount of the funds to be transferred.
        currency : str
            The ISO 4217 currency of the funds to be transferred.
        reference : str, optional
            The reference for the funds transfer.

        Returns
        -------
        dict | MoveMoneyBetweenAccounts.Response
            The details of the transfer.
        """
        endpoint = MoveMoneyBetweenAccounts
        path = endpoint.ROUTE
        body = endpoint.Body(
            request_id=request_id,
            source_account_id=source_account_id,
            target_account_id=target_account_id,
            amount=amount,
            currency=currency,
            reference=reference,
        )

        return self.client.post(
            path=path,
            response_model=endpoint.Response,
            body=body,
            **kwargs,
        )



from enum import StrEnum


class EnumWebhookEvent(StrEnum):
    """
    Enum for webhook event types
    """

    TRANSACTION_CREATED = "TransactionCreated"
    TRANSACTION_STATE_CHANGED = "TransactionStateChanged"
    PAYOUT_LINK_CREATED = "PayoutLinkCreated"
    PAYOUT_LINK_STATE_CHANGED = "PayoutLinkStateChanged"



from .common import PyRevolutBaseException


class PyRevolutServerUnavailable(PyRevolutBaseException):
    """Service Unavailable -- We're temporarily offline for maintenance. Please try again later."""

    def __init__(
        self,
        msg="The Revolut server is temporarily offline for maintenance. Please try again later.",
        *args,
        **kwargs,
    ):
        super().__init__(msg, *args, **kwargs)



from typing import Annotated

from pydantic import BaseModel, Field, model_validator

from pyrevolut.utils import DateTime
from pyrevolut.api.counterparties.resources import ResourceCounterparty


class RetrieveListOfCounterparties:
    """
    Get all the counterparties that you have created, or use the query parameters to filter the results.

    The counterparties are sorted by the created_at date in reverse chronological order.

    The returned counterparties are paginated. The maximum number of counterparties returned per page
    is specified by the limit parameter. To get to the next page, make a new request and use the
    created_at date of the last counterparty returned in the previous response.
    """

    ROUTE = "/1.0/counterparties"

    class Params(BaseModel):
        """
        Query parameters for the endpoint.
        """

        name: Annotated[
            str | None,
            Field(
                description="""
                The name of the counterparty to retrieve. It does not need to be an exact match, 
                partial match is also supported.
                """
            ),
        ] = None
        account_no: Annotated[
            str | None,
            Field(
                description="""
                The exact account number of the counterparty to retrieve.
                """
            ),
        ] = None
        sort_code: Annotated[
            str | None,
            Field(
                description="""
                The exact sort code of the counterparty to retrieve.
                Only allowed in combination with the account_no parameter.                
                """
            ),
        ] = None
        iban: Annotated[
            str | None,
            Field(
                description="""
                The exact IBAN of the counterparty to retrieve.
                """
            ),
        ] = None
        bic: Annotated[
            str | None,
            Field(
                description="""
                The exact BIC of the counterparty to retrieve. Only allowed in combination with the iban parameter.
                """
            ),
        ] = None
        created_before: Annotated[
            DateTime | None,
            Field(
                description="""
                Retrieves counterparties with created_at < created_before. 
                The default value is the current date and time at which you are calling the endpoint.
                Provided in ISO 8601 format.
                """
            ),
        ] = None
        limit: Annotated[
            int | None,
            Field(
                description="""
                The maximum number of counterparties returned per page.
                To get to the next page, make a new request and use the 
                created_at date of the last card returned in the previous 
                response as the value for created_before.              
                
                If not provided, the default value is 100.  
                """,
                ge=1,
                le=100,
            ),
        ] = None

        @model_validator(mode="after")
        def check_inputs(self) -> "RetrieveListOfCounterparties.Params":
            """
            Validate the input parameters.
            """
            if self.sort_code and not self.account_no:
                raise ValueError(
                    "The sort_code parameter is only allowed in combination with the account_no parameter."
                )
            if self.bic and not self.iban:
                raise ValueError(
                    "The bic parameter is only allowed in combination with the iban parameter."
                )
            return self

    class Response(ResourceCounterparty):
        """
        Response model for the endpoint.
        """

        pass



"""Module that contains the API classes."""

# flake8: noqa: F401, F403
from .accounts import *
from .cards import *
from .counterparties import *
from .foreign_exchange import *
from .payment_drafts import *
from .payout_links import *
from .simulations import *
from .team_members import *
from .transactions import *
from .transfers import *
from .webhooks import *
from .common import *



from typing import Annotated
from uuid import UUID

from pydantic import BaseModel, Field
from pydantic_extra_types.currency_code import Currency

from pyrevolut.utils import Date


class CreatePaymentDraft:
    """
    Create a payment draft.
    """

    ROUTE = "/1.0/payment-drafts"

    class Body(BaseModel):
        """
        The body model for the endpoint.
        """

        class ModelPayment(BaseModel):
            """The details of a payment"""

            class ModelReceiver(BaseModel):
                """The details of the transfer recipient.

                If the counterparty has multiple payment methods available
                (e.g. 2 accounts, or 1 account and 1 card), you must specify the
                account (account_id) or card (card_id) to which you want to transfer the money.
                """

                counterparty_id: Annotated[
                    UUID,
                    Field(description="The ID of the receiving counterparty."),
                ]
                account_id: Annotated[
                    UUID | None,
                    Field(
                        description="""
                        The ID of the receiving counterparty's account, which can be own account. 
                        Used for bank transfers.

                        If the counterparty has multiple payment methods available, use it to 
                        specify the account to which you want to send the money.
                        """
                    ),
                ] = None
                card_id: Annotated[
                    UUID | None,
                    Field(
                        description="""
                        The ID of the receiving counterparty's card. Used for card transfers.

                        If the counterparty has multiple payment methods available, use it to 
                        specify the card to which you want to send the money.
                        """
                    ),
                ] = None

            account_id: Annotated[
                UUID,
                Field(
                    description="""
                    The ID of the account to pay from.
                    You can specify only one account ID for multiple payments in the same payment draft.
                    """
                ),
            ]
            receiver: Annotated[
                ModelReceiver,
                Field(
                    description="""
                    The details of the transfer recipient.

                    If the counterparty has multiple payment methods available 
                    (e.g. 2 accounts, or 1 account and 1 card), you must specify the 
                    account (account_id) or card (card_id) to which you want to transfer the money.
                    """
                ),
            ]
            amount: Annotated[
                float,
                Field(description="The amount of the payment."),
            ]
            currency: Annotated[
                Currency,
                Field(description="ISO 4217 currency code in upper case."),
            ]
            reference: Annotated[
                str,
                Field(description="The reference for the payment."),
            ]

        title: Annotated[
            str | None,
            Field(description="The title of the payment draft."),
        ] = None
        schedule_for: Annotated[
            Date | None,
            Field(
                description="The scheduled date of the payment draft in ISO 8601 format."
            ),
        ] = None
        payments: Annotated[
            list[ModelPayment],
            Field(description="The details of the payments."),
        ]

    class Response(BaseModel):
        """
        The response model for the endpoint.
        """

        id: Annotated[UUID, Field(description="The ID of the payment draft created.")]



from typing import Annotated
from uuid import UUID

from pydantic import BaseModel, Field

from pyrevolut.utils import DateTime
from pyrevolut.api.common import EnumTransactionState


class SimulateTransferStateUpdate:
    """
    Simulate a transfer state change in the Sandbox environment.

    For example, after you make a transfer in Sandbox, you can change its
    state to completed.

    The resulting state is final and cannot be changed.
    """

    ROUTE = "/1.0/sandbox/transactions/{transfer_id}/{action}"

    class Body(BaseModel):
        """
        The body of the request.
        """

        pass

    class Response(BaseModel):
        """
        The response model.
        """

        id: Annotated[
            UUID, Field(description="The ID of the transfer whose state was updated.")
        ]
        state: Annotated[
            EnumTransactionState,
            Field(
                description="""
                Indicates the simulated transaction state. Possible values:

                completed:
                    Transaction was successfully processed.
                reverted:
                    Transaction was reverted by the system or company, but not the user. 
                    This can happen for a variety of reasons, for example, the receiver being inaccessible.
                declined:
                    Transaction was declined to the user for a good reason, such as insufficient 
                    account balance, wrong receiver information, etc.
                failed:
                    Transaction failed during initiation or completion. 
                    This can happen for a variety of reasons, for example, 
                    invalid API calls, blocked payments, etc.
                """
            ),
        ]
        created_at: Annotated[
            DateTime,
            Field(
                description="The date and time the transfer was created in ISO 8601 format."
            ),
        ]
        completed_at: Annotated[
            DateTime | None,
            Field(
                description="The date and time the transfer was completed in ISO 8601 format."
            ),
        ] = None



from typing import Annotated

from pydantic import BaseModel, Field
from pydantic_extra_types.country import CountryAlpha2

from pyrevolut.api.common import EnumPaymentScheme, EnumTimeUnit


class RetrieveFullBankDetails:
    """
    Get all the bank details of one of your accounts. Specify the account by its ID.
    """

    ROUTE = "/1.0/accounts/{account_id}/bank-details"

    class Params(BaseModel):
        """
        Query parameters for the endpoint.
        """

        pass

    class Response(BaseModel):
        """
        Response model for the endpoint.
        """

        class ModelBeneficiaryAddress(BaseModel):
            """The address of the beneficiary."""

            street_line1: Annotated[
                str | None,
                Field(description="Street line 1 information."),
            ] = None
            street_line2: Annotated[
                str | None,
                Field(description="Street line 2 information."),
            ] = None
            region: Annotated[
                str | None,
                Field(description="The name of the region."),
            ] = None
            city: Annotated[
                str | None,
                Field(description="The name of the city."),
            ] = None
            country: Annotated[
                CountryAlpha2,
                Field(
                    description="The country of the counterparty as the 2-letter ISO 3166 code."
                ),
            ]
            postcode: Annotated[
                str,
                Field(description="The postcode of the counterparty address."),
            ]

        class ModelEstimatedTime(BaseModel):
            """The estimated time of the inboud transfer of the funds,
            i.e. when we expect the recipient to receive the transfer.
            """

            unit: Annotated[
                EnumTimeUnit,
                Field(
                    description="The estimated time unit of the inbound transfer of the funds."
                ),
            ]
            min: Annotated[
                int | None,
                Field(description="The minimum time estimate.", ge=0),
            ] = None
            max: Annotated[
                int | None,
                Field(description="The maximum time estimate.", ge=0),
            ] = None

        iban: Annotated[
            str | None,
            Field(description="The IBAN number."),
        ] = None
        bic: Annotated[
            str | None,
            Field(description="The BIC number, also known as SWIFT code."),
        ] = None
        account_no: Annotated[
            str | None,
            Field(description="The account number."),
        ] = None
        sort_code: Annotated[
            str | None,
            Field(description="The sort code of the account."),
        ] = None
        routing_number: Annotated[
            str | None,
            Field(description="The routing number of the account."),
        ] = None
        beneficiary: Annotated[
            str,
            Field(description="The name of the counterparty."),
        ]
        beneficiary_address: Annotated[
            ModelBeneficiaryAddress,
            Field(description="The address of the counterparty."),
        ]
        bank_country: Annotated[
            CountryAlpha2 | None,
            Field(description="The country of the bank as the 2-letter ISO 3166 code."),
        ] = None
        pooled: Annotated[
            bool | None,
            Field(
                description="Indicates whether the account address is pooled or unique."
            ),
        ] = None
        unique_reference: Annotated[
            str | None,
            Field(description="The reference of the pooled account."),
        ] = None
        schemes: Annotated[
            list[EnumPaymentScheme],
            Field(
                description="The schemes that are available for this currency account."
            ),
        ]
        estimated_time: Annotated[
            ModelEstimatedTime,
            Field(
                description="""
                The estimated time of the inboud transfer of the funds, 
                i.e. when we expect the recipient to receive the transfer.
                """
            ),
        ]



"""Move funds in the same currency between accounts of your business,
or make payments to your counterparties.

Reference link: https://developer.revolut.com/docs/business/transfers
"""

# flake8: noqa: F401
from .endpoint import EndpointTransfersSync, EndpointTransfersAsync



from typing import Annotated

from pydantic import BaseModel, Field

from pyrevolut.utils import DateTime
from pyrevolut.api.common import EnumPayoutLinkState
from pyrevolut.api.payout_links.resources import ResourcePayoutLink


class RetrieveListOfPayoutLinks:
    """
    Get all the links that you have created, or use the query parameters to filter the results.

    The links are sorted by the created_at date in reverse chronological order.

    The returned links are paginated. The maximum number of payout links returned per
    page is specified by the limit parameter. To get to the next page, make a
    new request and use the created_at date of the last payout link returned in the previous response.

    Note
    ----
    This feature is available in the UK and the EEA.
    """

    ROUTE = "/1.0/payout-links"

    class Params(BaseModel):
        """
        The query parameters for the request.
        """

        state: Annotated[
            EnumPayoutLinkState | None,
            Field(
                description="""
                The state that the payout link is in. Possible states are:

                    created:
                        The payout link has been created, but the amount has not yet been blocked.
                    failed:
                        The payout link couldn't be generated due to a failure during transaction booking.
                    awaiting:
                        The payout link is awaiting approval.
                    active:
                        The payout link can be redeemed.
                    expired:
                        The payout link cannot be redeemed because it wasn't claimed before its expiry date.
                    cancelled:
                        The payout link cannot be redeemed because it was cancelled.
                    processing:
                        The payout link has been redeemed and is being processed.
                    processed:
                        The payout link has been redeemed and the money has been transferred to the recipient.
                """,
            ),
        ] = None
        created_before: Annotated[
            DateTime | None,
            Field(
                description="""
                Retrieves links with created_at < created_before. 
                The default value is the current date and time at which you are calling the endpoint.

                Provided in ISO 8601 format.
                """
            ),
        ] = None
        limit: Annotated[
            int | None,
            Field(
                description="""
                The maximum number of links returned per page.
                To get to the next page, make a new request and use the 
                created_at date of the last payout link returned in the previous 
                response as the value for created_before.              
                
                If not provided, the default value is 100.  
                """,
                ge=1,
                le=100,
            ),
        ] = None

    class Response(ResourcePayoutLink):
        """
        The response model for the request.
        """

        pass



"""This module contains the authentication methods."""

# flake8: noqa: F401
from .auth_manual import auth_manual_flow
from .creds import ModelCreds, load_creds, save_creds
from .enum_auth_scope import EnumAuthScope
from .get_auth_tokens import (
    ModelGetAuthTokensResponse,
    get_auth_tokens,
    aget_auth_tokens,
)
from .refresh_access_token import (
    ModelRefreshAccessTokenResponse,
    refresh_access_token,
    arefresh_access_token,
)



from .common import PyRevolutBaseException


class PyRevolutTooManyRequests(PyRevolutBaseException):
    """Too Many Requests -- You're sending too many requests."""

    def __init__(
        self,
        msg="You're sending too many requests. Please wait a while and try again.",
        *args,
        **kwargs,
    ):
        super().__init__(msg, *args, **kwargs)



from enum import StrEnum


class EnumAccountNameMatchCode(StrEnum):
    """Account Name Match Code Enum

    The result of the account name check. Possible values:

        matched:
            The name and account type match the provided values.
        close_match:
            The name and account type are similar to the provided values.
            The actual values are returned.
        not_matched:
            The name and account type don't match the provided values.
        cannot_be_checked:
            The check cannot be performed and retries won't help.
            For example, the recipient's bank doesn't support CoP.
        temporarily_unavailable:
            The check cannot be performed right now.
            For example, the recipient's bank didn't respond to our request.
            You should retry the request later.
    """

    MATCHED = "matched"
    CLOSE_MATCH = "close_match"
    NOT_MATCHED = "not_matched"
    CANNOT_BE_CHECKED = "cannot_be_checked"
    TEMPORARILY_UNAVAILABLE = "temporarily_unavailable"



from .common import PyRevolutBaseException


class PyRevolutInvalidEnvironment(PyRevolutBaseException):
    """Raised when the environment is invalid"""

    def __init__(
        self,
        msg="This request is not valid for the current environment. Please check the environment and try again.",
        *args,
        **kwargs,
    ):
        super().__init__(msg, *args, **kwargs)



import base64
import json

import pendulum
from httpx import Client
import typer
from rich import print as console

from pyrevolut.utils.datetime import to_datetime

from .enum_auth_scope import EnumAuthScope
from .creds import ModelCreds, save_creds
from .gen_public_private_cert import gen_public_private_cert
from .create_client_assert_jwt import create_client_assert_jwt
from .get_auth_tokens import get_auth_tokens


def auth_manual_flow(
    credentials_json: str = "credentials/creds.json",
    sandbox: bool = True,
    scopes: list[EnumAuthScope] | None = None,
):
    """
    Method to run the manual authorization flow to get the client assertion JWT and the access and refresh tokens.

    Parameters
    ----------
    credentials_json : str, optional
        The location to save the credentials JSON file.
        Default is "credentials/creds.json".
    sandbox : bool, optional
        Whether to use the sandbox environment.
        Default is True.
    scopes : list[EnumAuthScope] | None, optional
        The list of scopes to request. If not provided, the default scopes will be used.

        Access tokens can be issued with four security scopes and require a JWT (JSON Web Token)
        signature to be obtained:

            READ: Permissions for GET operations.

            WRITE: Permissions to update counterparties, webhooks, and issue payment drafts.

            PAY: Permissions to initiate or cancel transactions and currency exchanges.

            READ_SENSITIVE_CARD_DATA: Permissions to retrieve sensitive card details.

        Caution
        -------
        If you enable the READ_SENSITIVE_CARD_DATA scope for your access token, you must
        set up IP whitelisting.
        Failing to do so will prevent you from accessing any Business API endpoint.

        Default is None.

    Returns
    -------
    None
    """

    # Final credentials dictionary to be saved to a JSON file
    creds = {}

    console("======================================================================")
    console("=                    Revolut Client from Manual Flow                 =")
    console("======================================================================")

    # Check if the credentials file already exists
    try:
        with open(credentials_json, "r") as f:
            creds = json.load(f)
            console("Credentials file found. Loading credentials...")

        # Check if the credentials file is valid
        try:
            creds_model = ModelCreds(**creds)

            # Check for any expired fields
            if not creds_model.credentials_expired:
                console("Credentials loaded successfully.")
                console("To re-authenticate, please delete the credentials file.")
                return
            else:
                console("Credentials have expired. Re-authenticating...")
        except Exception as exc:
            console(f"An error occurred: {exc}")
            console(
                "Credentials file is invalid. Please delete the credentials file and re-authenticate."
            )
            return
    except FileNotFoundError:
        pass

    # Generate the Public and Private Certificates
    console("\n-----------------------------------------------------------")
    console("--- Step (1/4) Generate Public and Private Certificates ---")
    console("-----------------------------------------------------------")
    while True:
        expiration_dt = to_datetime(
            typer.prompt("Expiration datetime of the certificates (UTC)")
        )
        country = typer.prompt("Country (2-letter code)")
        email_address = typer.prompt(
            text="Email Address (can be left blank)", default=""
        )
        common_name = typer.prompt(
            text="Common Name (fully qualified host name, can be left blank)",
            default="",
        )
        state = typer.prompt(
            text="State or Province (full name, can be left blank)", default=""
        )
        locality = typer.prompt(text="Locality (city, can be left blank)", default="")
        organization = typer.prompt(
            text="Organization (company, can be left blank)", default=""
        )
        organization_unit = typer.prompt(
            text="Organization Unit (section, can be left blank)", default=""
        )

        try:
            keys = gen_public_private_cert(
                expiration_dt=expiration_dt,
                country=country,
                email_address=email_address,
                common_name=common_name,
                state=state,
                locality=locality,
                organization=organization,
                organization_unit=organization_unit,
                save_location_private=None,
                save_location_public=None,
            )
            public_key = keys["public"]
            private_key = keys["private"]

            # Store in the creds dictionary as base64 encoded strings
            creds["certificate"] = {
                "public": base64.b64encode(public_key).decode("utf-8"),
                "private": base64.b64encode(private_key).decode("utf-8"),
                "expiration_dt": expiration_dt.to_iso8601_string(),
            }
            break
        except Exception as exc:
            console(f"An error occurred: {exc}")

    console("\n-------------------------------------------------------")
    console("--- Step (2/4) Upload Public Certificate to Revolut ---")
    console("-------------------------------------------------------")
    upload_url = (
        "https://sandbox-business.revolut.com/settings/api"
        if sandbox
        else "https://business.revolut.com/settings/api"
    )
    public_key_string = public_key.decode("utf-8").replace("\\n", "")
    console(
        "Please specify your OAuth redirect URI. "
        "This is the URL where you are redirected after you consent for "
        "the application to access your Revolut Business account."
    )
    while True:
        try:
            redirect_url: str = typer.prompt(
                "OAuth redirect URI (e.g: https://example.com)",
                default="https://example.com",
            )
            if redirect_url.startswith("http://"):
                raise ValueError(
                    f"WARNING: Your redirect URL ({redirect_url}) is not secure."
                )
            break
        except ValueError:
            pass

    console("\nPlease follow the instructions exactly:")
    console(f"Step (2.1) Navigate to the following URL: {upload_url}")
    console(
        "Step (2.2) In the [bold red]API Certificates[/bold red] section, "
        "click [bold red]Add API certificate[/bold red]. "
        "If you already have other certificates added, click [bold red]Add new[/bold red]."
    )
    console(
        "Step (2.3) Give your certificate a meaningful title. It will help you later to distinguish "
        "it from other certificates."
    )
    console(
        "Step (2.4) Copy and paste the following url to the OAuth redirect URI field: "
        f"\n\n[bold]{redirect_url}[/bold]\n"
    )
    console(
        "Step (2.5) Copy and paste the following public key to the X509 public key field "
        "(including the -----BEGIN CERTIFICATE----- and -----END CERTIFICATE----- lines): "
        f"\n\n[bold]{public_key_string}[/bold]\n",
        end="",
    )
    console(
        "Step (2.6) Click [bold red]Continue[/bold red]. "
        "This takes you to the [bold red]API Certificate[/bold red] page with the parameters "
        "of your application."
    )
    console("Step (2.7) Copy the [bold red]Client ID[/bold red] provided by Revolut.")
    client_id = typer.prompt(text="Client ID")

    console("\n------------------------------------------------------")
    console("--- Step (3/4) Generate Client Assertion JWT Token ---")
    console("------------------------------------------------------")
    client_assert_jwt = create_client_assert_jwt(
        client_id=client_id,
        expiration_dt=expiration_dt,
        private_credentials_key=private_key,
        issuer_url=redirect_url,
        save_location=None,
    )
    creds["client_assert_jwt"] = {
        "jwt": client_assert_jwt,
        "expiration_dt": expiration_dt.to_iso8601_string(),
    }

    console("\n---------------------------------------------")
    console("--- Step (4/4) Consent to the application ---")
    console("---------------------------------------------")
    suffix = f"app-confirm?client_id={client_id}&redirect_uri={redirect_url}&response_type=code"
    if scopes is not None:
        suffix += f"&scope={','.join([scope for scope in scopes])}"
    suffix += "#authorize"
    consent_url = (
        "https://sandbox-business.revolut.com/"
        if sandbox
        else "https://business.revolut.com/"
    ) + suffix

    console("\nPlease follow the instructions exactly:")
    console(f"Step (4.1) Navigate to the following URL: {consent_url}")
    console(
        "Step (4.2) Click on [bold red]Authorize[/bold red] and upon success, "
        "you will be redirected to your OAuth redirect URI."
    )
    console("Step (4.3) Copy the entire redirect URI and paste it below.")
    auth_url: str = typer.prompt("OAuth redirect URI")
    auth_code = auth_url.split("code=")[1]

    console("\n-----------------------------------------------")
    console("--- Step (5/5) Get Access and Refresh Token ---")
    console("-----------------------------------------------")
    resp = get_auth_tokens(
        client=Client(),
        auth_code=auth_code,
        client_assert_jwt=client_assert_jwt,
        sandbox=sandbox,
    )
    dt_now = pendulum.now(tz="UTC")
    creds["tokens"] = {
        "access_token": resp.access_token.get_secret_value(),
        "refresh_token": resp.refresh_token.get_secret_value(),
        "token_type": resp.token_type,
        "access_token_expiration_dt": dt_now.add(
            seconds=resp.expires_in
        ).to_iso8601_string(),
        # To comply with PSD2 SCA regulations, refresh tokens expire after 90 days
        "refresh_token_expiration_dt": dt_now.add(days=90).to_iso8601_string(),
    }

    # Store the credentials to a JSON file
    creds_model = ModelCreds(**creds)
    save_creds(creds=creds_model, location=credentials_json, indent=4)

    console(
        "\n====================================================================================="
    )
    console(
        f"=                      Credentials Saved to {credentials_json}                  ="
    )
    console(
        "====================================================================================="
    )



from typing import Annotated

from pydantic import BaseModel, Field

from pyrevolut.utils import DateTime
from pyrevolut.api.common import EnumWebhookEvent

from .payout_link_created import ResourcePayoutLinkCreated
from .payout_link_state_changed import ResourcePayoutLinkStateChanged
from .transaction_created import ResourceTransactionCreated
from .transaction_state_changed import ResourceTransactionStateChanged


class ResourceWebhookPayload(BaseModel):
    """
    Webhook payload resource model.
    """

    event: Annotated[
        EnumWebhookEvent,
        Field(description="The event type."),
    ]
    timestamp: Annotated[
        DateTime,
        Field(description="The event time."),
    ]
    data: Annotated[
        ResourcePayoutLinkCreated
        | ResourcePayoutLinkStateChanged
        | ResourceTransactionCreated
        | ResourceTransactionStateChanged,
        Field(description="The event data."),
    ]



from pydantic import BaseModel

from pyrevolut.api.cards.resources import ResourceCard


class RetrieveCardDetails:
    """
    Get the details of a specific card, based on its ID.
    """

    ROUTE = "/1.0/cards/{card_id}"

    class Params(BaseModel):
        """
        Query parameters for the endpoint.
        """

        pass

    class Response(ResourceCard):
        """
        Response model for the endpoint.
        """

        pass



from typing import Type, TypeVar, Literal, Annotated
import logging
import json
import base64

from pydantic import BaseModel, Field
import pendulum

from httpx import (
    AsyncClient,
    Client as SyncClient,
    Request,
    Response,
    HTTPError,
    TimeoutException,
    NetworkError,
)

from pyrevolut.utils.auth import (
    ModelCreds,
    refresh_access_token,
    save_creds,
    load_creds,
)
from pyrevolut.exceptions import (
    PyRevolutBaseException,
    PyRevolutTimeoutError,
    PyRevolutNetworkError,
    PyRevolutBadRequest,
    PyRevolutUnauthorized,
    PyRevolutForbidden,
    PyRevolutNotFound,
    PyRevolutMethodNotAllowed,
    PyRevolutNotAcceptable,
    PyRevolutConflict,
    PyRevolutTooManyRequests,
    PyRevolutInternalServerError,
    PyRevolutServerUnavailable,
)


BM = TypeVar("BM", bound=Type[BaseModel])
D = TypeVar("D", dict, list)  # TypeVar for dictionary or list


class ModelError(BaseModel):
    """Model for the error response"""

    code: Annotated[int, Field(description="The error code")]
    message: Annotated[str, Field(description="The error message")]


class BaseClient:
    """The base client for the Revolut API"""

    creds_loc: str
    creds: str | dict | None = None
    credentials: ModelCreds
    domain: str
    sandbox: bool
    return_type: Literal["raw", "dict", "model"] = "dict"
    error_response: Literal["raw", "raise", "dict", "model"] = "raise"
    client: SyncClient | AsyncClient | None = None

    def __init__(
        self,
        creds_loc: str = "credentials/creds.json",
        creds: str | dict | None = None,
        sandbox: bool = True,
        return_type: Literal["raw", "dict", "model"] = "dict",
        error_response: Literal["raw", "raise", "dict", "model"] = "raise",
    ):
        """Create a new Revolut client

        Parameters
        ----------
        creds_loc : str, optional
            The location of the credentials file, by default "credentials/creds.json".
            If the creds input is not provided, will load the credentials from this file.
        creds : str | dict, optional
            The credentials to use for the client, by default None. If not provided, will
            load the credentials from the creds_loc file.
            Can be a dictionary of the credentials or a base64 encoded string of the credentials dictionary.
        sandbox : bool, optional
            Whether to use the sandbox environment, by default True
        return_type : Literal["raw", "dict", "model"], optional
            The return type for the API responses, by default "dict"
            If "raw":
                The raw response will be returned
            If "dict":
                The response will be the dictionary representation of the Pydantic model.
                So it will have UUIDs, pendulum DateTimes, etc instead of the raw string values.
            If "model":
                The response will be a Pydantic model containing all processed response data.
        error_response : Literal["raw", "raise", "dict", "model"], optional
            How the client should handle error responses, by default "raise"
            If "raw":
                The client will return the raw error response
            If "raise":
                The client will raise a ValueError if the response is an error
            If "dict":
                The client will return a dictionary representation of the error response
            If "model":
                The client will return a Pydantic model of the error response
        """
        assert return_type in [
            "raw",
            "dict",
            "model",
        ], "return_type must be 'raw', 'dict', or 'model'"
        assert error_response in [
            "raise",
            "dict",
            "model",
        ], "error_response must be 'raise', 'dict', or 'model'"

        self.creds_loc = creds_loc
        self.creds = creds
        self.sandbox = sandbox
        self.return_type = return_type
        self.error_response = error_response

        # Set domain based on environment
        if self.sandbox:
            self.domain = "https://sandbox-b2b.revolut.com/api"
        else:
            self.domain = "https://b2b.revolut.com/api"

        # Load the credentials
        self.load_credentials()

        # Load the endpoints
        self.load_endpoints()

    def process_response(
        self,
        response: Response,
        response_model: BM,
        return_type: Literal["raw", "dict", "model"] | None = None,
        error_response: Literal["raw", "raise", "dict", "model"] | None = "raise",
    ):
        """Processes the response and returns the desired format.
        Will additionally log the request and response.

        Parameters
        ----------
        response : Response
            The HTTPX response to process
        response_model : BM
            The Pydantic model to use for the response
        return_type : Literal["raw", "dict", "model"] | None, optional
            The return type for the API responses, by default None.
            If "raw":
                The raw response will be returned
            If "dict":
                The response will be the dictionary representation of the Pydantic model.
                So it will have UUIDs, pendulum DateTimes, etc instead of the raw string values.
            If "model":
                The response will be a Pydantic model containing all processed response data.
            If None:
                The default return type of the client will be used.
        error_response : Literal["raw", "raise", "dict", "model"] | None, optional
            How the client should handle error responses, by default None.
            If "raw":
                The client will return the raw error response
            If "raise":
                The client will raise a ValueError if the response is an error
            If "dict":
                The client will return a dictionary representation of the error response
            If "model":
                The client will return a Pydantic model of the error response
            If None:
                The default error response type of the client will be used.

        Returns
        -------
        BM | dict | list[BM] | list[dict]
            The response in the desired format
        """
        if return_type is None:
            return_type = self.return_type
        if error_response is None:
            error_response = self.error_response

        # Log the request
        self.log_request(request=response.request)

        # Log the response
        self.log_response(response=response)

        # Check for error response
        if response.is_error:
            if error_response == "raise":
                try:
                    response.raise_for_status()
                except TimeoutException as exc:
                    raise PyRevolutTimeoutError() from exc
                except NetworkError as exc:
                    raise PyRevolutNetworkError() from exc
                except HTTPError as exc:
                    if response.status_code == 400:
                        raise PyRevolutBadRequest() from exc
                    elif response.status_code == 401:
                        raise PyRevolutUnauthorized() from exc
                    elif response.status_code == 403:
                        raise PyRevolutForbidden() from exc
                    elif response.status_code == 404:
                        raise PyRevolutNotFound() from exc
                    elif response.status_code == 405:
                        raise PyRevolutMethodNotAllowed() from exc
                    elif response.status_code == 406:
                        raise PyRevolutNotAcceptable() from exc
                    elif response.status_code == 409:
                        raise PyRevolutConflict() from exc
                    elif response.status_code == 429:
                        raise PyRevolutTooManyRequests() from exc
                    elif response.status_code == 500:
                        raise PyRevolutInternalServerError() from exc
                    elif response.status_code == 503:
                        raise PyRevolutServerUnavailable() from exc
                    raise PyRevolutBaseException() from exc
            elif error_response == "raw":
                return response.json()
            elif error_response == "dict":
                return ModelError(**response.json()).model_dump()
            elif error_response == "model":
                return ModelError(**response.json())
            else:
                raise ValueError(f"Invalid error response type: {error_response}")

        # Raw response
        try:
            raw_response = response.json()
        except json.JSONDecodeError:
            raw_response = {}
        if return_type == "raw":
            return raw_response

        # Dict response
        if isinstance(raw_response, list):
            model_response = [response_model(**resp) for resp in raw_response]
        else:
            model_response = response_model(**raw_response)
        if return_type == "dict":
            if isinstance(model_response, list):
                return [resp.model_dump() for resp in model_response]
            return model_response.model_dump()

        # Model response
        if return_type == "model":
            return model_response

    def log_request(self, request: Request):
        """Log the request to the API

        Parameters
        ----------
        request : Request
            The request to log

        Returns
        -------
        None
        """
        logging.info(
            f"Request: {request.method} {request.url} - {request.headers} - {request.read().decode()}"
        )

    def log_response(self, response: Response):
        """Log the response from the API.

        Parameters
        ----------
        response : Response
            The response from the API

        Returns
        -------
        None
        """
        if not response.is_error:
            logging.info(f"Response: {response.status_code} - {response.text}")
        else:
            logging.error(f"Response: {response.status_code} - {response.text}")

    def _prep_get(
        self,
        path: str,
        params: Type[BaseModel] | None = None,
        **kwargs,
    ):
        """
        Method to prepare the GET request inputs for the HTTPX client.

        Parameters
        ----------
        path : str
            The path to send the request to
        params : Type[BaseModel] | None
            The parameters to add to the request route
        **kwargs
            Additional keyword arguments to pass to the HTTPX client

        Returns
        -------
        dict
            The prepared inputs for the HTTPX client
        """
        self.__check_client()
        path = self.__process_path(path)
        url = f"{self.domain}/{path}"
        headers = self.__create_headers(kwargs.pop("headers", {}))
        params = (
            self.__replace_null_with_none(
                data=params.model_dump(mode="json", exclude_none=True, by_alias=True)
            )
            if params is not None
            else None
        )
        return {
            "url": url,
            "params": params,
            "headers": headers,
            **kwargs,
        }

    def _prep_post(
        self,
        path: str,
        body: Type[BaseModel] | None = None,
        **kwargs,
    ):
        """
        Method to prepare the POST request inputs for the HTTPX client.

        Parameters
        ----------
        path : str
            The path to send the request to
        body : Type[BaseModel] | None
            The body to send in the request
        **kwargs
            Additional keyword arguments to pass to the HTTPX client

        Returns
        -------
        dict
            The prepared inputs for the HTTPX client
        """
        self.__check_client()
        path = self.__process_path(path)
        url = f"{self.domain}/{path}"
        headers = self.__create_headers(kwargs.pop("headers", {}))
        json = (
            self.__replace_null_with_none(
                data=body.model_dump(mode="json", exclude_none=True, by_alias=True)
            )
            if body is not None
            else None
        )
        return {
            "url": url,
            "json": json,
            "headers": headers,
            **kwargs,
        }

    def _prep_patch(
        self,
        path: str,
        body: Type[BaseModel] | None = None,
        **kwargs,
    ):
        """
        Method to prepare the PATCH request inputs for the HTTPX client.

        Parameters
        ----------
        path : str
            The path to send the request to
        body : Type[BaseModel] | None
            The body to send in the request
        **kwargs
            Additional keyword arguments to pass to the HTTPX client

        Returns
        -------
        dict
            The prepared inputs for the HTTPX client
        """
        self.__check_client()
        path = self.__process_path(path)
        url = f"{self.domain}/{path}"
        headers = self.__create_headers(kwargs.pop("headers", {}))
        json = (
            self.__replace_null_with_none(
                data=body.model_dump(mode="json", exclude_none=True, by_alias=True)
            )
            if body is not None
            else None
        )
        return {
            "url": url,
            "json": json,
            "headers": headers,
            **kwargs,
        }

    def _prep_delete(
        self,
        path: str,
        params: Type[BaseModel] | None = None,
        **kwargs,
    ):
        """
        Method to prepare the DELETE request inputs for the HTTPX client.

        Parameters
        ----------
        path : str
            The path to send the request to
        params : Type[BaseModel] | None
            The parameters to add to the request route
        **kwargs
            Additional keyword arguments to pass to the HTTPX client

        Returns
        -------
        dict
            The prepared inputs for the HTTPX client
        """
        self.__check_client()
        path = self.__process_path(path)
        url = f"{self.domain}/{path}"
        headers = self.__create_headers(kwargs.pop("headers", {}))
        params = (
            self.__replace_null_with_none(
                data=params.model_dump(mode="json", exclude_none=True, by_alias=True)
            )
            if params is not None
            else None
        )
        return {
            "url": url,
            "params": params,
            "headers": headers,
            **kwargs,
        }

    def _prep_put(
        self,
        path: str,
        body: Type[BaseModel] | None = None,
        **kwargs,
    ):
        """
        Method to prepare the PUT request inputs for the HTTPX client.

        Parameters
        ----------
        path : str
            The path to send the request to
        body : Type[BaseModel] | None
            The body to send in the request
        **kwargs
            Additional keyword arguments to pass to the HTTPX client

        Returns
        -------
        dict
            The prepared inputs for the HTTPX client
        """
        self.__check_client()
        path = self.__process_path(path)
        url = f"{self.domain}/{path}"
        headers = self.__create_headers(kwargs.pop("headers", {}))
        json = (
            self.__replace_null_with_none(
                data=body.model_dump(mode="json", exclude_none=True, by_alias=True)
            )
            if body is not None
            else None
        )
        return {
            "url": url,
            "json": json,
            "headers": headers,
            **kwargs,
        }

    def load_endpoints(self):
        """Loads all the endpoints from the api directory"""
        raise NotImplementedError("load_endpoints method must be implemented")

    @property
    def required_headers(self) -> dict[str, str]:
        """The headers to be attached to each request

        Returns
        -------
        dict[str, str]
            The headers to be attached to each request
        """
        return {
            "Accept": "application/json",
            "Authorization": f"Bearer {self.credentials.tokens.access_token.get_secret_value()}",
        }

    def __create_headers(self, headers: dict[str, str] = {}) -> dict[str, str]:
        """Create the headers for the request by adding the required headers

        Parameters
        ----------
        headers : dict[str, str]
            The headers for the request

        Returns
        -------
        dict[str, str]
            The headers for the request
        """
        headers.update(self.required_headers)
        return headers

    def __check_client(self):
        """Check if the client is open and that the credentials are still valid.

        Raises
        ------
        ValueError
            If the client is not open or if the long-term credentials have expired
        """
        if self.client is None:
            raise RuntimeError(
                "Client is not open. Use .open() or the contextmanager to open the client."
            )

        if self.credentials.credentials_expired:
            raise ValueError(
                "Long-term credentials have expired. "
                "\n\nPlease reauthenticate using the `pyrevolut auth-manual` command."
            )

        if self.credentials.access_token_expired:
            self.refresh_access_token()

    def __process_path(self, path: str) -> str:
        """Process the path.

        If 'http' not in the path:
            Removing the leading slash if it exists
        Else:
            Return the path as is

        Parameters
        ----------
        path : str
            The path to process

        Returns
        -------
        str
            The processed path
        """
        if "http" in path:
            return path

        return self.__remove_leading_slash(path=path)

    def __remove_leading_slash(self, path: str) -> str:
        """Remove the leading slash from a path if it exists and
        return it without the leading slash

        Parameters
        ----------
        path : str
            The path to remove the leading slash from

        Returns
        -------
        str
            The path without the leading slash
        """
        if path.startswith("/"):
            return path[1:]
        return path

    def __replace_null_with_none(self, data: D) -> D:
        """
        Method that replaces all 'null' strings with None in a provided dictionary or list.

        Must be called with either a dictionary or a list, not both.

        Parameters
        ----------
        data : dict | list
            The dictionary or list to replace 'null' strings with None

        Returns
        -------
        dict | list
            The dictionary or list with 'null' strings replaced with None
        """
        if isinstance(data, dict):
            for k, v in data.items():
                if isinstance(v, dict):
                    self.__replace_null_with_none(data=v)
                elif isinstance(v, list):
                    self.__replace_null_with_none(data=v)
                elif v == "null":
                    data[k] = None
        elif isinstance(data, list):
            for i in range(len(data)):
                if isinstance(data[i], dict):
                    self.__replace_null_with_none(data=data[i])
                elif isinstance(data[i], list):
                    self.__replace_null_with_none(data=data[i])
                elif data[i] == "null":
                    data[i] = None
        else:
            raise ValueError("Data must be either a dictionary or a list")

        return data

    def load_credentials(self):
        """Load the credentials from the credentials inputs.

        - If credentials are not provided, will load them from the credentials file.
        - If the credentials file does not exist, raise an error.
        - If the credentials file is invalid, raise an error.
        - If the credentials are expired, raise an error.
        - If the access token is expired, refresh it.

        """
        solution_msg = (
            "\n\nPlease reauthenticate using the `pyrevolut auth-manual` command."
        )

        if self.creds is not None:
            if isinstance(self.creds, str):
                _creds = json.loads(base64.b64decode(self.creds).decode("utf-8"))
            else:
                _creds = self.creds
            try:
                self.credentials = ModelCreds(**_creds)
            except Exception as exc:
                raise ValueError(
                    f"Error loading credentials: {exc}. {solution_msg}"
                ) from exc
        else:
            try:
                self.credentials = load_creds(location=self.creds_loc)
            except FileNotFoundError as exc:
                raise ValueError(
                    f"Credentials file not found: {exc}. {solution_msg}"
                ) from exc
            except Exception as exc:
                raise ValueError(
                    f"Error loading credentials: {exc}. {solution_msg}"
                ) from exc

        # Check if the credentials are still valid
        if self.credentials.credentials_expired:
            raise ValueError(f"Credentials are expired. {solution_msg}")

        # Check if the access token is expired
        if self.credentials.access_token_expired:
            self.refresh_access_token()

    def refresh_access_token(self):
        """Refresh the access token using the refresh token.
        Will call the endpoint to refresh the access token.
        Then it will save the new access token to the credentials file.

        Parameters
        ----------
        None

        Raises
        ------
        ValueError
            If there is an error refreshing the access token.

        Returns
        -------
        None
        """
        try:
            resp = refresh_access_token(
                client=SyncClient(),
                refresh_token=self.credentials.tokens.refresh_token.get_secret_value(),
                client_assert_jwt=self.credentials.client_assert_jwt.jwt.get_secret_value(),
                sandbox=self.sandbox,
            )
            self.credentials.tokens.access_token = resp.access_token.get_secret_value()
            self.credentials.tokens.token_type = resp.token_type
            self.credentials.tokens.access_token_expiration_dt = pendulum.now(
                tz="UTC"
            ).add(seconds=resp.expires_in)
            save_creds(creds=self.credentials, location=self.creds_loc, indent=4)
        except Exception as exc:
            raise ValueError(f"Error refreshing access token: {exc}.") from exc



"""This module holds the Webhooks resource models."""

# flake8: noqa: F401

from .payout_link_created import ResourcePayoutLinkCreated
from .payout_link_state_changed import ResourcePayoutLinkStateChanged
from .transaction_created import ResourceTransactionCreated
from .transaction_state_changed import ResourceTransactionStateChanged
from .webhook_payload import ResourceWebhookPayload
from .webhook import ResourceWebhook
